from itertools import chain
import unittest
from unittest.mock import Mock
from queue import Queue

from AnyQt.QtGui import QStandardItem

from Orange.data import Table
from Orange.widgets.visualize.utils import (
    VizRankDialog, Result, run_vizrank, QueuedScore
)
from Orange.widgets.tests.base import WidgetTest


def compute_score(x):
    return (x[0] + 1) / (x[1] + 1)


class TestRunner(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data = Table("iris")

    def test_Result(self):
        res = Result(queue=Queue(), scores=[])
        self.assertIsInstance(res.queue, Queue)
        self.assertIsInstance(res.scores, list)

    def test_run_vizrank(self):
        scores, task = [], Mock()
        # run through all states
        task.is_interruption_requested.return_value = False
        states = [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)]
        res = run_vizrank(compute_score, lambda initial: chain(states),
                          None, scores, 0, 6, task)

        next_state = self.assertQueueEqual(
            res.queue, [0, 0, 0, 3, 2, 5], compute_score,
            states, states[1:] + [None])
        self.assertIsNone(next_state)
        res_scores = sorted([compute_score(x) for x in states])
        self.assertListEqual(res.scores, res_scores)
        self.assertIsNot(scores, res.scores)
        self.assertEqual(task.set_partial_result.call_count, 2)
        self.assertEqual(task.set_progress_value.call_count, 7)

    def test_run_vizrank_interrupt(self):
        scores, task = [], Mock()
        # interrupt calculation in third iteration
        task.is_interruption_requested.side_effect = lambda: \
            True if task.is_interruption_requested.call_count > 2 else False
        states = [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)]
        res = run_vizrank(compute_score, lambda initial: chain(states),
                          None, scores, 0, 6, task)

        next_state = self.assertQueueEqual(
            res.queue, [0, 0], compute_score, states[:2], states[1:3])
        self.assertEqual(next_state, (0, 3))
        res_scores = sorted([compute_score(x) for x in states[:2]])
        self.assertListEqual(res.scores, res_scores)
        self.assertIsNot(scores, res.scores)
        self.assertEqual(task.set_partial_result.call_count, 1)
        self.assertEqual(task.set_progress_value.call_count, 3)
        task.set_progress_value.assert_called_with(int(1 / 6 * 100))

        # continue calculation through all states
        task.is_interruption_requested.side_effect = lambda: False
        i = states.index(next_state)
        res = run_vizrank(compute_score, lambda initial: chain(states[i:]),
                          None, res_scores, 2, 6, task)

        next_state = self.assertQueueEqual(
            res.queue, [0, 3, 2, 5], compute_score, states[2:],
            states[3:] + [None])
        self.assertIsNone(next_state)
        res_scores = sorted([compute_score(x) for x in states])
        self.assertListEqual(res.scores, res_scores)
        self.assertIsNot(scores, res.scores)
        self.assertEqual(task.set_partial_result.call_count, 3)
        self.assertEqual(task.set_progress_value.call_count, 8)
        task.set_progress_value.assert_called_with(int(5 / 6 * 100))

    def assertQueueEqual(self, queue, positions, f, states, next_states):
        self.assertIsInstance(queue, Queue)
        for qs in (QueuedScore(position=p, score=f(s), state=s, next_state=ns)
                   for p, s, ns in zip(positions, states, next_states)):
            result = queue.get_nowait()
            self.assertEqual(result.position, qs.position)
            self.assertEqual(result.state, qs.state)
            self.assertEqual(result.next_state, qs.next_state)
            self.assertEqual(result.score, qs.score)
            next_state = result.next_state
        return next_state


class TestVizRankDialog(WidgetTest):
    def test_on_partial_result(self):
        def iterate_states(initial_state):
            if initial_state is not None:
                return chain(states[states.index(initial_state):])
            return chain(states)

        def invoke_on_partial_result():
            widget.on_partial_result(run_vizrank(
                widget.compute_score,
                widget.iterate_states,
                widget.saved_state,
                widget.scores,
                widget.saved_progress,
                widget.state_count(),
                task
            ))

        task = Mock()
        states = [(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)]

        widget = VizRankDialog(None)
        widget.progressBarInit()
        widget.compute_score = compute_score
        widget.iterate_states = iterate_states
        widget.row_for_state = lambda sc, _: [QStandardItem(str(sc))]
        widget.state_count = lambda: len(states)

        # interrupt calculation in third iteration
        task.is_interruption_requested.side_effect = lambda: \
            True if task.is_interruption_requested.call_count > 2 else False
        invoke_on_partial_result()
        self.assertEqual(widget.rank_model.rowCount(), 2)
        for row, score in enumerate(
                sorted([compute_score(x) for x in states[:2]])):
            self.assertEqual(widget.rank_model.item(row, 0).text(), str(score))
        self.assertEqual(widget.saved_progress, 2)
        task.set_progress_value.assert_called_with(int(1 / 6 * 100))

        # continue calculation through all states
        task.is_interruption_requested.side_effect = lambda: False
        invoke_on_partial_result()
        self.assertEqual(widget.rank_model.rowCount(), 6)
        for row, score in enumerate(
                sorted([compute_score(x) for x in states])):
            self.assertEqual(widget.rank_model.item(row, 0).text(), str(score))
        self.assertEqual(widget.saved_progress, 6)
        task.set_progress_value.assert_called_with(int(5 / 6 * 100))


if __name__ == "__main__":
    unittest.main()



from unittest.mock import patch

import random
import unittest

import numpy as np

from Orange.data import Table, Domain, StringVariable, DiscreteVariable, \
    ContinuousVariable
from Orange.data.filter import SameValue
from Orange.widgets.utils.annotated_data import (
    create_annotated_table, create_groups_table, ANNOTATED_DATA_FEATURE_NAME,
    lazy_annotated_table, lazy_groups_table, domain_with_annotation_column
)


class TestAnnotatedData(unittest.TestCase):
    def setUp(self):
        random.seed(42)
        self.zoo = Table("zoo")

    def test_domain_with_annotation_column(self):
        a, b, c = (ContinuousVariable(x) for x in "abc")

        x = [[1, 2, 3], [4, 5, 6]]

        for data in (dabc := Domain([a, b, c]), Table.from_list(dabc, x)):
            dom, var = domain_with_annotation_column(data)
            self.assertEqual(dom.attributes, (a, b, c))
            self.assertIs(dom.class_var, var)
            self.assertEqual(var.name, ANNOTATED_DATA_FEATURE_NAME)
            self.assertEqual(var.values, ("No", "Yes"))

            dom, var = domain_with_annotation_column(
                data, values=tuple("xyz"), var_name="d")
            self.assertEqual(dom.attributes, (a, b, c))
            self.assertIs(dom.class_var, var)
            self.assertEqual(var.name, "d")
            self.assertEqual(var.values, tuple("xyz"))

        for data in (dabc := Domain([a, b], c), Table.from_list(dabc, x)):
            dom, var = domain_with_annotation_column(
                data, values=tuple("xyz"), var_name="d")
            self.assertEqual(dom.attributes, (a, b))
            self.assertIs(dom.class_var, c)
            self.assertEqual(dom.metas, (var, ))
            self.assertEqual(var.name, "d")
            self.assertEqual(var.values, tuple("xyz"))

            dom, var = domain_with_annotation_column(
                data, values=tuple("xyz"), var_name="c")
            self.assertEqual(dom.attributes, (a, b))
            self.assertIs(dom.class_var, c)
            self.assertEqual(dom.metas, (var, ))
            self.assertEqual(var.name, "c (1)")
            self.assertEqual(var.values, tuple("xyz"))

    def test_create_annotated_table(self):
        annotated = create_annotated_table(self.zoo, list(range(10)))

        # check annotated table domain
        self.assertEqual(annotated.domain.variables, self.zoo.domain.variables)
        self.assertEqual(2, len(annotated.domain.metas))
        self.assertIn(self.zoo.domain.metas[0], annotated.domain.metas)
        self.assertIn(ANNOTATED_DATA_FEATURE_NAME,
                      [m.name for m in annotated.domain.metas])

        # check annotated table data
        np.testing.assert_array_equal(annotated.X, self.zoo.X)
        np.testing.assert_array_equal(annotated.Y, self.zoo.Y)
        np.testing.assert_array_equal(annotated.metas[:, 0].ravel(),
                                      self.zoo.metas.ravel())
        self.assertEqual(
            10, np.sum([i[ANNOTATED_DATA_FEATURE_NAME] for i in annotated]))

    def test_create_annotated_table_selected(self):
        # check annotated column for no selected indices
        annotated = create_annotated_table(self.zoo, [])
        self.assertEqual(len(annotated), len(self.zoo))
        self.assertEqual(
            0, np.sum([i[ANNOTATED_DATA_FEATURE_NAME] for i in annotated]))

        # check annotated column fol all selectes indices
        annotated = create_annotated_table(self.zoo, list(range(len(self.zoo))))
        self.assertEqual(len(annotated), len(self.zoo))
        self.assertEqual(
            len(self.zoo),
            np.sum([i[ANNOTATED_DATA_FEATURE_NAME] for i in annotated]))

    def test_create_annotated_table_none_data(self):
        self.assertIsNone(create_annotated_table(None, None))

    def test_create_annotated_table_none_indices(self):
        annotated = create_annotated_table(self.zoo, None)
        self.assertEqual(len(annotated), len(self.zoo))
        self.assertEqual(
            0, np.sum([i[ANNOTATED_DATA_FEATURE_NAME] for i in annotated]))

    def _renamed_zoo_meta(self, name, name0=None):
        zoo = self.zoo
        zood = self.zoo.domain
        var0 = zood.attributes[0]
        name0 = name0 or var0.name
        domain = Domain(
            (DiscreteVariable(name0, var0.values), ) + zood.attributes[1:],
            zood.class_var, [StringVariable(name)])
        return Table(domain, zoo.X, zoo.Y, zoo.metas)

    def test_cascade_annotated_tables(self):
        # check cascade of annotated tables
        data = self._renamed_zoo_meta(ANNOTATED_DATA_FEATURE_NAME)
        first_meta = data.domain.metas[0]
        for i in range(5):
            data = create_annotated_table(
                data, random.sample(range(0, len(data)), 20))
            self.assertEqual(2 + i, len(data.domain.metas))
            self.assertIn(first_meta, data.domain.metas)
            self.assertIn(ANNOTATED_DATA_FEATURE_NAME,
                          [m.name for m in data.domain.metas])
            for j in range(1, i + 2):
                self.assertIn("{} ({})".format(ANNOTATED_DATA_FEATURE_NAME, j),
                              [m.name for m in data.domain.metas])

    def test_cascade_annotated_tables_with_missing_middle_feature(self):
        # check table for domain [..., "Feature", "Selected", "Selected (3)] ->
        # [..., "Feature", "Selected", "Selected (3), "Selected (4)"]
        data = self._renamed_zoo_meta(f"{ANNOTATED_DATA_FEATURE_NAME} (3)",
                                      ANNOTATED_DATA_FEATURE_NAME)
        data = create_annotated_table(
            data, random.sample(range(0, len(self.zoo)), 20))
        self.assertEqual(2, len(data.domain.metas))
        self.assertEqual(data.domain.attributes[0].name,
                         ANNOTATED_DATA_FEATURE_NAME)
        self.assertEqual(data.domain.metas[0].name,
                         "{} ({})".format(ANNOTATED_DATA_FEATURE_NAME, 3))
        self.assertEqual(data.domain.metas[1].name,
                         "{} ({})".format(ANNOTATED_DATA_FEATURE_NAME, 4))

    def test_cascade_annotated_tables_with_missing_annotated_feature(self):
        # check table for domain [..., "Feature", "Selected (3)] ->
        # [..., "Feature", "Selected (3), "Selected (4)"]
        data = self._renamed_zoo_meta(f"{ANNOTATED_DATA_FEATURE_NAME} (3)")
        data = create_annotated_table(
            data, random.sample(range(0, len(self.zoo)), 20))
        self.assertEqual(2, len(data.domain.metas))
        self.assertEqual(data.domain.metas[0].name,
                         "{} ({})".format(ANNOTATED_DATA_FEATURE_NAME, 3))
        self.assertEqual(data.domain.metas[1].name,
                         "{} ({})".format(ANNOTATED_DATA_FEATURE_NAME, 4))

    def test_create_groups_table_include_unselected(self):
        group_indices = random.sample(range(0, len(self.zoo)), 20)
        selection = np.zeros(len(self.zoo), dtype=np.uint8)
        selection[group_indices[:10]] = 1
        selection[group_indices[10:]] = 2
        table = create_groups_table(self.zoo, selection)
        selvar = table.domain["Selected"]
        self.assertEqual(
            len(SameValue(selvar, "Unselected")(table)),
            len(self.zoo) - len(group_indices)
        )
        self.assertEqual(selvar.values, ("G1", "G2", "Unselected"))

    def test_create_groups_table_set_values(self):
        group_indices = random.sample(range(0, len(self.zoo)), 20)
        selection = np.zeros(len(self.zoo), dtype=np.uint8)
        selection[group_indices[:10]] = 1
        selection[group_indices[10:]] = 2
        values = ("this", "that", "rest")
        table = create_groups_table(self.zoo, selection, values=values)
        self.assertEqual(tuple(table.domain["Selected"].values), values)

    @patch("Orange.widgets.utils.annotated_data.create_annotated_table")
    def test_lazy_annotated_table(self, creator):
        selected_indices = np.array([1, 2, 3])
        lazy_table = lazy_annotated_table(self.zoo, selected_indices)
        self.assertEqual(lazy_table.length, len(self.zoo))
        self.assertEqual(lazy_table.domain.attributes, self.zoo.domain.attributes)
        self.assertEqual(lazy_table.domain.class_var, self.zoo.domain.class_var)
        self.assertEqual(len(lazy_table.domain.metas), 2)
        var = lazy_table.domain.metas[1]
        self.assertIsInstance(var, DiscreteVariable)
        self.assertEqual(var.name, ANNOTATED_DATA_FEATURE_NAME)
        creator.assert_not_called()
        self.assertIs(lazy_table.get_value(), creator.return_value)

    @patch("Orange.widgets.utils.annotated_data.create_groups_table")
    def test_lazy_groups_table(self, creator):
        group_indices = np.zeros(len(self.zoo), dtype=int)
        group_indices[10:15] = 1

        lazy_table = lazy_groups_table(self.zoo, group_indices)
        self.assertEqual(lazy_table.length, len(self.zoo))
        self.assertEqual(lazy_table.domain.attributes, self.zoo.domain.attributes)
        self.assertEqual(lazy_table.domain.class_var, self.zoo.domain.class_var)
        self.assertEqual(len(lazy_table.domain.metas), 2)
        var = lazy_table.domain.metas[1]
        self.assertIsInstance(var, DiscreteVariable)
        self.assertEqual(var.name, ANNOTATED_DATA_FEATURE_NAME)
        creator.assert_not_called()
        self.assertIs(lazy_table.get_value(), creator.return_value)
        creator.reset_mock()

        lazy_table = lazy_groups_table(
            self.zoo, group_indices, include_unselected=False, var_name="foo",
            values=("bar", "baz"))
        self.assertEqual(lazy_table.length, 5)
        self.assertEqual(lazy_table.domain.attributes, self.zoo.domain.attributes)
        self.assertEqual(lazy_table.domain.class_var, self.zoo.domain.class_var)
        self.assertEqual(len(lazy_table.domain.metas), 2)
        var = lazy_table.domain.metas[1]
        self.assertIsInstance(var, DiscreteVariable)
        self.assertEqual(var.name, "foo")
        self.assertEqual(var.values, ("bar", "baz"))
        creator.assert_not_called()
        self.assertIs(lazy_table.get_value(), creator.return_value)


if __name__ == "__main__":
    unittest.main()



from os.path import dirname
import unittest

import Orange


def suite(loader=None, pattern='test*.py'):
    top_level_dir = dirname(dirname(Orange.__file__))
    return unittest.TestSuite(loader.discover(dirname(__file__),
                                              pattern or 'test*.py',
                                              top_level_dir))


def load_tests(loader, tests, pattern):
    return suite(loader, pattern)


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



#pylint: disable=missing-docstring,protected-access,unsubscriptable-object
import unittest
from unittest.mock import Mock

import numpy as np

from Orange.data import Table, Domain, ContinuousVariable
from Orange.widgets.data.owneighbors import OWNeighbors, METRICS
from Orange.widgets.tests.base import WidgetTest, ParameterMapping


class TestOWNeighbors(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWNeighbors,
                                         stored_settings={"auto_apply": False})
        self.iris = Table("iris")

    def test_input_data(self):
        """Check widget's data with data on the input"""
        widget = self.widget
        self.assertEqual(widget.data, None)
        self.send_signal(widget.Inputs.data, self.iris)
        self.assertEqual(widget.data, self.iris)

    def test_input_data_disconnect(self):
        """Check widget's data after disconnecting data on the input"""
        widget = self.widget
        self.send_signal(widget.Inputs.data, self.iris)
        self.assertEqual(widget.data, self.iris)
        self.send_signal(widget.Inputs.data, None)
        self.assertEqual(widget.data, None)

    def test_input_reference(self):
        """Check widget's reference with reference on the input"""
        widget = self.widget
        self.assertEqual(widget.reference, None)
        self.send_signal(widget.Inputs.reference, self.iris)
        self.assertEqual(widget.reference, self.iris)

    def test_input_reference_disconnect(self):
        """Check reference after disconnecting reference on the input"""
        widget = self.widget
        self.send_signal(widget.Inputs.data, self.iris)
        self.send_signal(widget.Inputs.reference, self.iris)
        self.assertEqual(widget.reference, self.iris)
        self.send_signal(widget.Inputs.reference, None)
        self.assertEqual(widget.reference, None)
        widget.apply_button.button.click()
        self.assertIsNone(self.get_output())

    def test_output_neighbors(self):
        """Check if neighbors are on the output after apply"""
        widget = self.widget
        self.assertIsNone(self.get_output())
        self.send_signals(((widget.Inputs.data, self.iris),
                           (widget.Inputs.reference, self.iris[:10])))
        widget.apply_button.button.click()
        self.assertIsNotNone(self.get_output())
        self.assertIsInstance(self.get_output(), Table)
        self.assertTrue(all([i in self.iris.ids for i in
                             self.get_output(widget.Outputs.data).ids])
                        )

    def test_settings(self):
        """Check neighbors for various distance metrics"""
        widget = self.widget
        settings = [
            ParameterMapping("", widget.controls.distance_index, METRICS),
            ParameterMapping("", widget.controls.n_neighbors)]
        for setting in settings:
            for val in setting.values:
                self.send_signal(widget.Inputs.data, self.iris)
                self.send_signal(widget.Inputs.reference, self.iris[:10])
                setting.set_value(val)
                widget.apply_button.button.click()
                if METRICS[widget.distance_index][0] != "Jaccard" \
                        and widget.n_neighbors != 0:
                    self.assertIsNotNone(self.get_output())

    def test_exclude_reference(self):
        """Check neighbors when reference is excluded"""
        widget = self.widget
        reference = self.iris[:5]
        self.send_signal(widget.Inputs.data, self.iris)
        self.send_signal(widget.Inputs.reference, reference)
        self.widget.exclude_reference = True
        widget.apply_button.button.click()
        neighbors = self.get_output(widget.Outputs.data)
        for inst in reference:
            self.assertNotIn(inst, neighbors)

    def test_similarity(self):
        widget = self.widget
        reference = self.iris[:10]
        self.send_signal(widget.Inputs.data, self.iris)
        self.send_signal(widget.Inputs.reference, reference)
        widget.apply_button.button.click()
        neighbors = self.get_output()
        self.assertEqual(self.iris.domain.attributes,
                         neighbors.domain.attributes)
        self.assertEqual(self.iris.domain.class_vars,
                         neighbors.domain.class_vars)
        self.assertIn("distance", neighbors.domain)
        self.assertTrue(all(100 >= ins["distance"] >= 0 for ins in neighbors))

    def test_missing_values(self):
        widget = self.widget
        data = Table("iris")
        reference = data[:3]
        with data.unlocked():
            data.X[0:10, 0] = np.nan
        self.send_signal(widget.Inputs.data, self.iris)
        self.send_signal(widget.Inputs.reference, reference)
        widget.apply_button.button.click()
        self.assertIsNotNone(self.get_output())

    def test_compute_distances_apply_called(self):
        """Check compute distances and apply are called when receiving signal"""
        widget = self.widget
        cdist = widget.compute_distances = Mock()
        def_commit = widget.commit.now = Mock()
        self.widget.auto_apply = False
        data = Table("iris")
        self.send_signal(widget.Inputs.data, data)
        cdist.assert_called()
        def_commit.assert_called()
        cdist.reset_mock()
        def_commit.reset_mock()

        self.send_signal(widget.Inputs.reference, data[:10])
        cdist.assert_called()
        def_commit.assert_called()
        cdist.reset_mock()
        def_commit.reset_mock()

        self.send_signals([(widget.Inputs.data, data),
                           (widget.Inputs.reference, data[:10])])
        self.assertEqual(cdist.call_count, 1)
        self.assertEqual(def_commit.call_count, 1)

    def test_compute_distances_calls_distance(self):
        widget = self.widget
        widget.distance_index = 2
        dists = np.random.random((10, 5))
        distance = Mock(return_value=dists)
        try:
            orig_metrics = METRICS[widget.distance_index]
            METRICS[widget.distance_index] = ("foo", distance)
            data = Table("iris")
            data, refs = data[:10], data[-5:]
            self.send_signals([(widget.Inputs.data, data),
                               (widget.Inputs.reference, refs)])
            # false positive, pylint: disable=unsubscriptable-object
            d1, d2 = distance.call_args[0]
            np.testing.assert_almost_equal(d1.X, data.X)
            np.testing.assert_almost_equal(d2.X, refs.X)
            # false positive, pylint: disable=no-member
            np.testing.assert_almost_equal(widget.distances, dists.min(axis=1))
        finally:
            METRICS[widget.distance_index] = orig_metrics

    def test_compute_distances_distance_no_data(self):
        widget = self.widget
        distance = Mock()
        try:
            orig_metrics = METRICS[widget.distance_index]
            METRICS[widget.distance_index] = ("foo", distance)
            data = Table("iris")
            data, refs = data[:10], data[-5:]
            self.send_signals([(widget.Inputs.data, data),
                               (widget.Inputs.reference, None)])
            distance.assert_not_called()
            self.assertIsNone(widget.distances)
            self.send_signal(widget.Inputs.data, data)
            distance.assert_not_called()
            self.assertIsNone(widget.distances)
            self.send_signals([(widget.Inputs.data, None),
                               (widget.Inputs.reference, refs)])
            distance.assert_not_called()
            self.assertIsNone(widget.distances)
            self.send_signals([(widget.Inputs.data, data),
                               (widget.Inputs.reference, None)])
            distance.assert_not_called()
            self.assertIsNone(widget.distances)

            data, refs = data[:0], data[:0]
            self.send_signals([(widget.Inputs.data, data),
                               (widget.Inputs.reference, None)])
            distance.assert_not_called()
            self.assertIsNone(widget.distances)
        finally:
            METRICS[widget.distance_index] = orig_metrics

    def test_compute_indices_without_reference(self):
        widget = self.widget
        widget.limit_neighbours = True

        # Indices for easier reading: 0 1  2  3  4  5  6  7  8  9 10 11 12
        widget.distances = np.array([4, 1, 7, 0, 5, 2, 4, 0, 2, 2, 2, 9, 8])

        widget.data = Mock()
        widget.data.ids = np.arange(13)
        widget.reference = Mock()
        widget.reference.ids = np.array([1, 3])

        widget.n_neighbors = 5
        self.assertEqual(sorted(widget._compute_indices()), [5, 7, 8, 9, 10])

        widget.n_neighbors = 1
        self.assertEqual(list(widget._compute_indices()), [7])

        widget.n_neighbors = 3
        ind = set(widget._compute_indices())
        self.assertEqual(len(ind), 3)
        self.assertIn(7, ind)
        self.assertTrue(len({5, 8, 9, 10} & ind) == 2)

        widget.n_neighbors = 100
        self.assertEqual(
            sorted(widget._compute_indices()),
            [0, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12])

        widget.n_neighbors = 10
        self.assertEqual(
            sorted(widget._compute_indices()),
            [0, 2, 4, 5, 6, 7, 8, 9, 10, 12])

        widget.limit_neighbours = False
        self.assertEqual(
            sorted(widget._compute_indices()),
            [0, 2, 4, 5, 6, 7, 8, 9, 10, 12])

    def test_data_with_similarity(self):
        widget = self.widget
        indices = np.array([5, 10, 15, 100])

        data = Table("iris")
        widget.data = data

        widget.distances = np.arange(1000, 1150).astype(float)
        neighbours = widget._data_with_similarity(indices)
        self.assertEqual(neighbours.metas.shape, (4, 1))
        np.testing.assert_almost_equal(
            neighbours.metas.flatten(), indices + 1000)
        np.testing.assert_almost_equal(neighbours.X, data.X[indices])

        domain = data.domain
        domain2 = Domain([domain[2]], domain.class_var, metas=domain[:2])
        data2 = data.transform(domain2)
        widget.data = data2

        widget.distances = np.arange(1000, 1150).astype(float)
        neighbours = widget._data_with_similarity(indices)
        self.assertEqual(len(neighbours.domain.metas), 3)
        self.assertEqual(neighbours.metas.shape, (4, 3))
        np.testing.assert_almost_equal(
            neighbours.get_column("distance"), indices + 1000)
        np.testing.assert_almost_equal(neighbours.X, data2.X[indices])

    def test_apply(self):
        widget = self.widget
        widget.auto_apply = True
        data = Table("iris")
        indices = np.array([5, 10, 15, 100])

        widget._compute_indices = lambda: \
            indices if widget.distances is not None else None
        self.send_signal(widget.Inputs.data, data)
        self.send_signal(widget.Inputs.reference, data[42:43])
        neigh = self.get_output(widget.Outputs.data)
        np.testing.assert_almost_equal(neigh.X, data.X[indices])
        np.testing.assert_almost_equal(
            neigh.metas.flatten(), widget.distances[indices])

    def test_all_equal_ref(self):
        widget = self.widget
        widget.auto_apply = True

        data = Table("iris")
        self.send_signal(widget.Inputs.data, data[:10])
        self.send_signal(widget.Inputs.reference, data[:10])
        self.assertTrue(widget.Warning.all_data_as_reference.is_shown())
        self.assertFalse(widget.Info.removed_references.is_shown())
        self.assertIsNone(self.get_output(widget.Outputs.data))

        self.send_signal(widget.Inputs.data, data[:15])
        self.assertFalse(widget.Warning.all_data_as_reference.is_shown())
        self.assertTrue(widget.Info.removed_references.is_shown())
        self.assertIsNotNone(self.get_output(widget.Outputs.data))

    def test_different_domains(self):
        """
        Test weather widget show error when data and a reference have different
        domains.
        """
        w = self.widget

        domain = Domain([ContinuousVariable("a")])
        domain_ref = Domain([ContinuousVariable("b")])
        data = Table(domain, np.random.rand(2, len(domain.variables)))
        reference = Table(
            domain_ref, np.random.rand(1, len(domain_ref.variables)))

        # no error if one or both of the signals is missing
        self.send_signal(w.Inputs.data, data)
        self.assertFalse(w.Error.diff_domains.is_shown())

        self.send_signal(w.Inputs.data, None)
        self.assertFalse(w.Error.diff_domains.is_shown())

        self.send_signal(w.Inputs.reference, data[:1])
        self.assertFalse(w.Error.diff_domains.is_shown())

        # same domain - no error
        self.send_signal(w.Inputs.data, data)
        self.send_signal(w.Inputs.reference, data[:1])
        self.assertFalse(w.Error.diff_domains.is_shown())

        # one attribute different attribute name
        self.send_signal(w.Inputs.data, data)
        self.send_signal(w.Inputs.reference, reference)
        self.assertTrue(w.Error.diff_domains.is_shown())

        # different number of attributes
        domain_ref = Domain([ContinuousVariable("a"), ContinuousVariable("b")])
        reference = Table(domain_ref, np.random.rand(1, len(domain_ref.variables)))

        # error disappears when data is set to None
        self.send_signal(w.Inputs.data, data)
        self.send_signal(w.Inputs.reference, reference)
        self.assertTrue(w.Error.diff_domains.is_shown())

        self.send_signal(w.Inputs.data, None)
        self.assertFalse(w.Error.diff_domains.is_shown())

        # error disappears when reference is set to None
        self.send_signal(w.Inputs.data, data)
        self.send_signal(w.Inputs.reference, reference)
        self.assertTrue(w.Error.diff_domains.is_shown())

        self.send_signal(w.Inputs.reference, None)
        self.assertFalse(w.Error.diff_domains.is_shown())

    def test_different_metas(self):
        """
        Test weather widget do not show error when data and a reference have
        domain that differ only in metas
        """
        w = self.widget

        domain = Domain([ContinuousVariable("a"), ContinuousVariable("b")],
                        metas=[ContinuousVariable("c")])
        data = Table(
            domain, np.random.rand(15, len(domain.attributes)),
            metas=np.random.rand(15, len(domain.metas)))

        # same domain with same metas no error
        self.send_signal(w.Inputs.data, data)
        self.send_signal(w.Inputs.reference, data[:1])
        self.assertFalse(w.Error.diff_domains.is_shown())
        output = self.get_output(w.Outputs.data)
        self.assertEqual(10, len(output))

        # same domain with different metas no error
        domain_ref = Domain(domain.attributes,
                            metas=[ContinuousVariable("d")])
        reference = Table(
            domain_ref, np.random.rand(1, len(domain_ref.attributes)),
            metas=np.random.rand(1, len(domain.metas)))
        self.send_signal(w.Inputs.data, data)
        self.send_signal(w.Inputs.reference, reference)
        self.assertFalse(w.Error.diff_domains.is_shown())
        output = self.get_output(w.Outputs.data)
        self.assertEqual(10, len(output))

        # same domain with different order - no error
        domain_ref = Domain(domain.attributes[::-1])
        reference = Table(
            domain_ref, np.random.rand(1, len(domain_ref.attributes)))
        self.send_signal(w.Inputs.data, data)
        self.send_signal(w.Inputs.reference, reference)
        self.assertFalse(w.Error.diff_domains.is_shown())
        output = self.get_output(w.Outputs.data)
        self.assertEqual(10, len(output))

        # same domain with different number of metas no error
        domain_ref = Domain(
            domain.attributes,
            metas=[ContinuousVariable("d"), ContinuousVariable("e")])
        reference = Table(
            domain_ref, np.random.rand(1, len(domain_ref.attributes)),
            metas=np.random.rand(1, len(domain_ref.metas)))
        self.send_signal(w.Inputs.data, data)
        self.send_signal(w.Inputs.reference, reference)
        self.assertFalse(w.Error.diff_domains.is_shown())
        output = self.get_output(w.Outputs.data)
        self.assertEqual(10, len(output))

        # different domain with same metas - error shown
        domain_ref = Domain(domain.attributes + (ContinuousVariable("e"),),
                            metas=[ContinuousVariable("c")])
        reference = Table(
            domain_ref, np.random.rand(1, len(domain_ref.attributes)),
            metas=np.random.rand(1, len(domain_ref.metas)))
        self.send_signal(w.Inputs.data, data)
        self.send_signal(w.Inputs.reference, reference)
        self.assertTrue(w.Error.diff_domains.is_shown())
        output = self.get_output(w.Outputs.data)
        self.assertIsNone(output)

    def test_different_domains_same_names(self):
        """
        Here we create two domains with same names of attributes - the case
        with image-analytics (two embeddings). Widget must not show the
        error in this case.
        """
        w = self.widget
        domain1 = Domain([ContinuousVariable("a"), ContinuousVariable("b")],
                         metas=[ContinuousVariable("c")])
        domain2 = Domain([ContinuousVariable("a"), ContinuousVariable("b")],
                         metas=[ContinuousVariable("d")])
        data = Table(
            domain1, np.random.rand(15, len(domain1.attributes)),
            metas=np.random.rand(15, len(domain1.metas)))
        ref = Table(
            domain2, np.random.rand(2, len(domain2.attributes)),
            metas=np.random.rand(2, len(domain2.metas)))
        self.send_signal(w.Inputs.data, data)
        self.send_signal(w.Inputs.reference, ref)
        self.assertFalse(w.Error.diff_domains.is_shown())
        output = self.get_output(w.Outputs.data)
        self.assertEqual(10, len(output))

    def test_n_neighbours_spin_max(self):
        w = self.widget
        sb = w.controls.n_neighbors
        default = sb.maximum()
        self.send_signal(w.Inputs.data, self.iris)
        self.assertEqual(sb.maximum(), len(self.iris))
        self.send_signal(w.Inputs.data, self.iris[:20])
        self.assertEqual(sb.maximum(), 20)
        self.send_signal(w.Inputs.data, None)
        self.assertEqual(sb.maximum(), default)

    def test_inherited_table(self):
        # pylint: disable=abstract-method
        class Table2(Table):
            pass

        data = Table2(self.iris)
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.reference, data[0:1])
        self.assertIsInstance(self.get_output(self.widget.Outputs.data), Table2)

    def test_order_by_distance(self):
        domain = Domain([ContinuousVariable(x) for x in "ab"])
        reference = Table.from_numpy(domain, [[1, 0]])
        data = Table.from_numpy(domain, [[1, 0.1], [2, 0], [1, 0], [0, 0.1], [0.1, 0]])
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.reference, reference)

        output = self.get_output(self.widget.Outputs.data)
        expected = [[1, 0], [1, 0.1], [0.1, 0], [2, 0], [0, 0.1]]
        np.testing.assert_array_equal(output.X, expected)
        dst = output.get_column("distance").tolist()
        self.assertTrue(dst == sorted(dst))  # check distance in ascending order

        # test on bigger set
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.send_signal(self.widget.Inputs.reference, self.iris[:1])
        dst = self.get_output(self.widget.Outputs.data).get_column("distance").tolist()
        self.assertTrue(dst == sorted(dst))  # check distance in ascending order


if __name__ == "__main__":
    unittest.main()



# pylint: disable=all
import numpy as np

from AnyQt.QtCore import Qt, QPoint
from AnyQt.QtGui import QPalette, QColor
from AnyQt.QtTest import QTest
from AnyQt.QtWidgets import QGraphicsScene, QGraphicsView

from orangewidget.tests.utils import mouseMove
from orangewidget.tests.base import GuiTest

from Orange.clustering import hierarchical
from Orange.widgets.utils.dendrogram import DendrogramWidget


T = hierarchical.Tree
C = hierarchical.ClusterData
S = hierarchical.SingletonData


def t(h: float, left: T, right: T):
    return T(C((left.value.first, right.value.last), h), (left, right))


def leaf(r, index):
    return T(S((r, r + 1), 0.0, index))


class TestDendrogramWidget(GuiTest):
    def setUp(self) -> None:
        super().setUp()
        self.scene = QGraphicsScene()
        self.view = QGraphicsView(self.scene)
        self.view.resize(300, 300)
        self.widget = DendrogramWidget()
        self.scene.addItem(self.widget)

    def tearDown(self) -> None:
        self.scene.clear()
        del self.widget
        super().tearDown()

    def test_widget(self):
        w = self.widget
        w.set_root(t(0.0, leaf(0, 0), leaf(1, 1)))
        w.resize(w.effectiveSizeHint(Qt.PreferredSize))
        h = w.height_at(QPoint())
        self.assertEqual(h, 0)
        h = w.height_at(QPoint(10, 0))
        self.assertEqual(h, 0)

        self.assertEqual(w.pos_at_height(0).x(), w.rect().x())
        self.assertEqual(w.pos_at_height(1).x(), w.rect().x())

        height = np.finfo(float).eps
        w.set_root(t(height, leaf(0, 0), leaf(1, 1)))

        h = w.height_at(QPoint())
        self.assertEqual(h, height)
        h = w.height_at(QPoint(int(w.size().width()), 0))
        self.assertEqual(h, 0)

        self.assertEqual(w.pos_at_height(0).x(), w.rect().right())
        self.assertEqual(w.pos_at_height(height).x(), w.rect().left())

        view = self.view
        view.grab()  # ensure w is laid out
        root = w.root()
        rootitem = w.item(root)
        r = view.mapFromScene(rootitem.sceneBoundingRect()).boundingRect()
        # move/hover over the item
        mouseMove(view.viewport(), r.center())
        self.assertEqual(w._highlighted_item, rootitem)
        # click select
        QTest.mouseClick(view.viewport(), Qt.LeftButton, Qt.NoModifier, r.center())
        self.assertTrue(w.isItemSelected(rootitem))
        p = r.topLeft() + QPoint(-3, -3)  # just out of the item
        mouseMove(view.viewport(), p)
        self.assertEqual(w._highlighted_item, None)

    def test_update_palette(self):
        w = self.widget
        w.set_root(t(1.0, leaf(0, 0), leaf(1, 1)))
        w.setSelectedClusters([w.root()])
        p = QPalette()
        p.setColor(QPalette.All, QPalette.WindowText, QColor(Qt.red))
        w.setPalette(p)
        item = w.item(w.root())
        self.assertEqual(item.pen().color(), p.color(QPalette.WindowText))



import os
import logging
from itertools import chain
from urllib.parse import urlparse
from typing import List, Dict, Any

import numpy as np
from AnyQt.QtWidgets import \
    QStyle, QComboBox, QMessageBox, QGridLayout, QLabel, \
    QLineEdit, QSizePolicy as Policy, QCompleter
from AnyQt.QtCore import Qt, QTimer, QSize, QUrl
from AnyQt.QtGui import QBrush

from orangewidget.utils.filedialogs import format_filter
from orangewidget.workflow.drophandler import SingleUrlDropHandler

from Orange.data.table import Table, get_sample_datasets_dir
from Orange.data.io import FileFormat, UrlReader, class_from_qualified_name
from Orange.data.io_base import MissingReaderException
from Orange.util import log_warnings
from Orange.widgets import widget, gui
from Orange.widgets.utils.localization import pl
from Orange.widgets.settings import Setting, ContextSetting, \
    PerfectDomainContextHandler, SettingProvider
from Orange.widgets.utils.domaineditor import DomainEditor
from Orange.widgets.utils.itemmodels import PyListModel
from Orange.widgets.utils.filedialogs import RecentPathsWComboMixin, \
    open_filename_dialog, stored_recent_paths_prepend
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import Output, Msg
from Orange.widgets.utils.combobox import TextEditCombo
from Orange.widgets.utils.state_summary import missing_values


# Backward compatibility: class RecentPath used to be defined in this module,
# and it is used in saved (pickled) settings. It must be imported into the
# module's namespace so that old saved settings still work
from Orange.widgets.utils.filedialogs import RecentPath

DEFAULT_READER_TEXT = "Automatically detect type"

log = logging.getLogger(__name__)


def add_origin(examples, filename):
    """
    Adds attribute with file location to each string variable
    Used for relative filenames stored in string variables (e.g. pictures)
    TODO: we should consider a cleaner solution (special variable type, ...)
    """
    if not filename:
        return
    strings = [var
               for var in examples.domain.variables + examples.domain.metas
               if var.is_string]
    dir_name, _ = os.path.split(filename)
    for var in strings:
        if "type" in var.attributes and "origin" not in var.attributes:
            var.attributes["origin"] = dir_name


class NamedURLModel(PyListModel):
    def __init__(self, mapping):
        self.mapping = mapping
        super().__init__()

    def data(self, index, role=Qt.DisplayRole):
        data = super().data(index, role)
        if role == Qt.DisplayRole:
            return self.mapping.get(data, data)
        return data

    def add_name(self, url, name):
        self.mapping[url] = name
        self.modelReset.emit()


class LineEditSelectOnFocus(QLineEdit):
    def focusInEvent(self, event):
        super().focusInEvent(event)
        # If selectAll is called directly, placing the cursor unselects the text
        QTimer.singleShot(0, self.selectAll)


class OWFile(widget.OWWidget, RecentPathsWComboMixin):
    name = "File"
    id = "orange.widgets.data.file"
    description = "Read data from an input file or network " \
                  "and send a data table to the output."
    icon = "icons/File.svg"
    priority = 10
    category = "Data"
    keywords = "file, load, read, open"

    class Outputs:
        data = Output("Data", Table,
                      doc="Attribute-valued dataset read from the input file.")

    want_main_area = False
    buttons_area_orientation = None

    SEARCH_PATHS = [("sample-datasets", get_sample_datasets_dir())]
    SIZE_LIMIT = 1e7
    LOCAL_FILE, URL = range(2)

    settingsHandler = PerfectDomainContextHandler(
        match_values=PerfectDomainContextHandler.MATCH_VALUES_ALL
    )

    # pylint seems to want declarations separated from definitions
    recent_paths: List[RecentPath]
    recent_urls: List[str]
    variables: list

    # Overload RecentPathsWidgetMixin.recent_paths to set defaults
    recent_paths = Setting([
        RecentPath("", "sample-datasets", "iris.tab"),
        RecentPath("", "sample-datasets", "titanic.tab"),
        RecentPath("", "sample-datasets", "housing.tab"),
        RecentPath("", "sample-datasets", "heart_disease.tab"),
        RecentPath("", "sample-datasets", "brown-selected.tab"),
        RecentPath("", "sample-datasets", "zoo.tab"),
    ])
    recent_urls = Setting([])
    source = Setting(LOCAL_FILE)
    sheet_names = Setting({})
    url = Setting("")

    variables = ContextSetting([])

    domain_editor = SettingProvider(DomainEditor)

    class Information(widget.OWWidget.Information):
        no_file_selected = Msg("No file selected.")

    class Warning(widget.OWWidget.Warning):
        file_too_big = Msg("The file is too large to load automatically."
                           " Press Reload to load.")
        load_warning = Msg("Read warning:\n{}")
        performance_warning = Msg(
            "Categorical variables with >100 values may decrease performance.")
        renamed_vars = Msg("Some variables have been renamed "
                           "to avoid duplicates.\n{}")
        multiple_targets = Msg("Most widgets do not support multiple targets")

    class Error(widget.OWWidget.Error):
        file_not_found = Msg("File not found.")
        missing_reader = Msg("Missing reader.")
        sheet_error = Msg("Error listing available sheets.")
        unknown = Msg("Read error:\n{}")

    UserAdviceMessages = [
        widget.Message(
            "Use CSV File Import widget for advanced options "
            "for comma-separated files",
            "use-csv-file-import"),
        widget.Message(
            "This widget loads only tabular data. Use other widgets to load "
            "other data types like models, distance matrices and networks.",
            "other-data-types"
        )
    ]

    def __init__(self):
        super().__init__()
        RecentPathsWComboMixin.__init__(self)
        self.domain = None
        self.data = None
        self.loaded_file = ""
        self.reader = None

        readers = [f for f in FileFormat.formats
                   if getattr(f, 'read', None)
                   and getattr(f, "EXTENSIONS", None)]

        def group_readers_per_addon_key(w):
            # readers from Orange.data.io should go first
            def package(w):
                package = w.qualified_name().split(".")[:-1]
                package = package[:2]
                if ".".join(package) == "Orange.data":
                    return ["0"]  # force "Orange" to come first
                return package
            return package(w), w.DESCRIPTION

        self.available_readers = sorted(set(readers),
                                        key=group_readers_per_addon_key)

        layout = QGridLayout()
        layout.setSpacing(4)
        gui.widgetBox(self.controlArea, orientation=layout, box='Source')
        vbox = gui.radioButtons(None, self, "source", box=True,
                                callback=self.load_data, addToLayout=False)

        rb_button = gui.appendRadioButton(vbox, "File:", addToLayout=False)
        layout.addWidget(rb_button, 0, 0, Qt.AlignVCenter)

        box = gui.hBox(None, addToLayout=False, margin=0)
        box.setSizePolicy(Policy.Expanding, Policy.Fixed)
        self.file_combo.setSizePolicy(Policy.Expanding, Policy.Fixed)
        self.file_combo.setMinimumSize(QSize(100, 1))
        self.file_combo.activated[int].connect(self.select_file)
        box.layout().addWidget(self.file_combo)
        layout.addWidget(box, 0, 1)

        file_button = gui.button(
            None, self, '...', callback=self.browse_file, autoDefault=False)
        file_button.setIcon(self.style().standardIcon(QStyle.SP_DirOpenIcon))
        file_button.setSizePolicy(Policy.Maximum, Policy.Fixed)
        layout.addWidget(file_button, 0, 2)

        reload_button = gui.button(
            None, self, "Reload", callback=self.load_data, autoDefault=False)
        reload_button.setIcon(self.style().standardIcon(
            QStyle.SP_BrowserReload))
        reload_button.setSizePolicy(Policy.Fixed, Policy.Fixed)
        layout.addWidget(reload_button, 0, 3)

        self.sheet_box = gui.hBox(None, addToLayout=False, margin=0)
        self.sheet_combo = QComboBox()
        self.sheet_combo.textActivated.connect(self.select_sheet)
        self.sheet_combo.setSizePolicy(Policy.Expanding, Policy.Fixed)
        self.sheet_combo.setMinimumSize(QSize(50, 1))
        self.sheet_label = QLabel()
        self.sheet_label.setText('Sheet')
        self.sheet_label.setSizePolicy(
            Policy.MinimumExpanding, Policy.Fixed)
        self.sheet_box.layout().addWidget(
            self.sheet_label, Qt.AlignLeft)
        self.sheet_box.layout().addWidget(
            self.sheet_combo, Qt.AlignVCenter)
        layout.addWidget(self.sheet_box, 2, 1)
        self.sheet_box.hide()

        rb_button = gui.appendRadioButton(vbox, "URL:", addToLayout=False)
        layout.addWidget(rb_button, 3, 0, Qt.AlignVCenter)

        self.url_combo = url_combo = TextEditCombo()
        url_model = NamedURLModel(self.sheet_names)
        url_model.wrap(self.recent_urls)
        url_combo.setLineEdit(LineEditSelectOnFocus())
        url_combo.setModel(url_model)
        url_combo.setSizePolicy(Policy.Ignored, Policy.Fixed)
        url_combo.setInsertPolicy(url_combo.InsertAtTop)
        url_edit = url_combo.lineEdit()
        margins = url_edit.textMargins()
        l, t, r, b = margins.left(), margins.top(), margins.right(), margins.bottom()
        url_edit.setTextMargins(l + 5, t, r, b)
        layout.addWidget(url_combo, 3, 1, 1, 3)
        url_combo.activated.connect(self._url_set)
        # whit completer we set that combo box is case sensitive when
        # matching the history
        completer = QCompleter()
        completer.setCaseSensitivity(Qt.CaseSensitive)
        url_combo.setCompleter(completer)

        layout = QGridLayout()
        layout.setSpacing(4)
        gui.widgetBox(self.controlArea, orientation=layout, box='File Type')

        box = gui.hBox(None, addToLayout=False, margin=0)
        box.setSizePolicy(Policy.Expanding, Policy.Fixed)
        self.reader_combo = QComboBox(self)
        self.reader_combo.setSizePolicy(Policy.Expanding, Policy.Fixed)
        self.reader_combo.setMinimumSize(QSize(100, 1))
        self.reader_combo.activated[int].connect(self.select_reader)

        box.layout().addWidget(self.reader_combo)
        layout.addWidget(box, 0, 1)

        box = gui.vBox(self.controlArea, "Info")
        self.infolabel = gui.widgetLabel(box, 'No data loaded.')

        box = gui.widgetBox(self.controlArea, "Columns (Double click to edit)")
        self.domain_editor = DomainEditor(self)
        self.editor_model = self.domain_editor.model()
        box.layout().addWidget(self.domain_editor)

        box = gui.hBox(box)
        gui.button(
            box, self, "Reset", callback=self.reset_domain_edit,
            autoDefault=False
        )
        gui.rubber(box)
        self.apply_button = gui.button(
            box, self, "Apply", callback=self.apply_domain_edit)
        self.apply_button.setEnabled(False)
        self.apply_button.setFixedWidth(170)
        self.editor_model.dataChanged.connect(
            lambda: self.apply_button.setEnabled(True))

        hBox = gui.hBox(self.controlArea)
        gui.rubber(hBox)
        gui.button(
            hBox, self, "Browse documentation datasets",
            callback=lambda: self.browse_file(True), autoDefault=False)
        gui.rubber(hBox)

        self.set_file_list()
        # Must not call open_file from within __init__. open_file
        # explicitly re-enters the event loop (by a progress bar)

        self.setAcceptDrops(True)

        if self.source == self.LOCAL_FILE:
            last_path = self.last_path()
            if last_path and os.path.exists(last_path) and \
                    os.path.getsize(last_path) > self.SIZE_LIMIT:
                self.Warning.file_too_big()
                return

        QTimer.singleShot(0, self.load_data)

    @staticmethod
    def sizeHint():
        return QSize(600, 550)

    def select_file(self, n):
        assert n < len(self.recent_paths)
        super().select_file(n)
        if self.recent_paths:
            self.source = self.LOCAL_FILE
            self.load_data()
            self.set_file_list()

    def select_sheet(self):
        self.recent_paths[0].sheet = self.sheet_combo.currentText()
        self.load_data()

    def select_reader(self, n):
        if self.source != self.LOCAL_FILE:
            return  # ignore for URL's

        if self.recent_paths:
            path = self.recent_paths[0]
            if n == 0:  # default
                path.file_format = None
                self.load_data()
            elif n <= len(self.available_readers):
                reader = self.available_readers[n - 1]
                path.file_format = reader.qualified_name()
                self.load_data()
            else:  # the rest include just qualified names
                path.file_format = self.reader_combo.itemText(n)
                self.load_data()

    def _url_set(self):
        index = self.url_combo.currentIndex()
        url = self.url_combo.currentText()
        url = url.strip()

        if not urlparse(url).scheme:
            url = 'http://' + url
            self.url_combo.setItemText(index, url)

        if index != 0:
            model = self.url_combo.model()
            root = self.url_combo.rootModelIndex()
            model.moveRow(root, index, root, 0)
            assert self.url_combo.currentIndex() == 0

        self.source = self.URL
        self.load_data()

    def browse_file(self, in_demos=False):
        if in_demos:
            start_file = get_sample_datasets_dir()
            if not os.path.exists(start_file):
                QMessageBox.information(
                    None, "File",
                    "Cannot find the directory with documentation datasets")
                return
        else:
            start_file = self.last_path() or os.path.expanduser("~/")

        filename, reader, _ = open_filename_dialog(start_file, None, self.available_readers)
        if not filename:
            return
        self.add_path(filename)
        if reader is not None:
            self.recent_paths[0].file_format = reader.qualified_name()

        self.source = self.LOCAL_FILE
        self.load_data()

    # Open a file, create data from it and send it over the data channel
    def load_data(self):
        # We need to catch any exception type since anything can happen in
        # file readers
        self.closeContext()
        self.domain_editor.set_domain(None)
        self.apply_button.setEnabled(False)
        self.clear_messages()
        self.set_file_list()

        error = self._try_load()
        if error:
            error()
            self.data = None
            self.sheet_box.hide()
            self.Outputs.data.send(None)
            self.infolabel.setText("No data.")

    def _try_load(self):
        self._initialize_reader_combo()

        # pylint: disable=broad-except
        if self.source == self.LOCAL_FILE:
            if self.last_path() is None:
                return self.Information.no_file_selected
            elif not os.path.exists(self.last_path()):
                return self.Error.file_not_found
        else:
            url = self.url_combo.currentText().strip()
            if not url:
                return self.Information.no_file_selected

        def mark_problematic_reader():
            self.reader_combo.setItemData(self.reader_combo.currentIndex(),
                                          QBrush(Qt.red), Qt.ForegroundRole)

        try:
            self.reader = self._get_reader()  # also sets current reader index
            assert self.reader is not None
        except MissingReaderException:
            mark_problematic_reader()
            return self.Error.missing_reader
        except Exception as ex:
            mark_problematic_reader()
            log.exception(ex)
            return lambda x=ex: self.Error.unknown(str(x))

        try:
            self._update_sheet_combo()
        except Exception:
            return self.Error.sheet_error

        with log_warnings() as warnings:
            try:
                data = self.reader.read()
            except Exception as ex:
                mark_problematic_reader()
                log.exception(ex)
                return lambda x=ex: self.Error.unknown(str(x))
            if warnings:
                self.Warning.load_warning(warnings[-1].message.args[0])

        self.infolabel.setText(self._describe(data))

        self.loaded_file = self.last_path()
        add_origin(data, self.loaded_file)
        self.data = data
        self.openContext(data.domain)
        self.apply_domain_edit()  # sends data
        return None

    def _get_reader(self) -> FileFormat:
        if self.source == self.LOCAL_FILE:
            path = self.last_path()
            self.reader_combo.setEnabled(True)
            if self.recent_paths and self.recent_paths[0].file_format:
                qname = self.recent_paths[0].file_format
                qname_index = {r.qualified_name(): i for i, r in enumerate(self.available_readers)}
                if qname in qname_index:
                    self.reader_combo.setCurrentIndex(qname_index[qname] + 1)
                else:
                    # reader may be accessible, but not in self.available_readers
                    # (perhaps its code was moved)
                    self.reader_combo.addItem(qname)
                    self.reader_combo.setCurrentIndex(len(self.reader_combo) - 1)
                try:
                    reader_class = class_from_qualified_name(qname)
                except Exception as ex:
                    raise MissingReaderException(f'Can not find reader "{qname}"') from ex
                reader = reader_class(path)
            else:
                self.reader_combo.setCurrentIndex(0)
                reader = FileFormat.get_reader(path)
            if self.recent_paths and self.recent_paths[0].sheet:
                reader.select_sheet(self.recent_paths[0].sheet)
            return reader
        else:
            url = self.url_combo.currentText().strip()
            return UrlReader(url)

    def _update_sheet_combo(self):
        if len(self.reader.sheets) < 2:
            self.sheet_box.hide()
            self.reader.select_sheet(None)
            return

        self.sheet_combo.clear()
        self.sheet_combo.addItems(self.reader.sheets)
        self._select_active_sheet()
        self.sheet_box.show()

    def _select_active_sheet(self):
        try:
            idx = self.reader.sheets.index(self.reader.sheet)
            self.sheet_combo.setCurrentIndex(idx)
        except ValueError:
            # Requested sheet does not exist in this file
            self.reader.select_sheet(None)
            self.sheet_combo.setCurrentIndex(0)

    def _initialize_reader_combo(self):
        self.reader_combo.clear()
        filters = [format_filter(f) for f in self.available_readers]
        self.reader_combo.addItems([DEFAULT_READER_TEXT] + filters)
        self.reader_combo.setCurrentIndex(0)
        self.reader_combo.setDisabled(True)
        # additional readers may be added in self._get_reader()

    @staticmethod
    def _describe(table):
        domain = table.domain
        text = ""

        attrs = getattr(table, "attributes", {})
        descs = [attrs[desc]
                 for desc in ("Name", "Description") if desc in attrs]
        if len(descs) == 2:
            descs[0] = f"<b>{descs[0]}</b>"
        if descs:
            text += f"<p>{'<br/>'.join(descs)}</p>"

        text += f"<p>{len(table)} {pl(len(table), 'instance')}"

        missing_in_attr = missing_in_class = ""
        if table.X.size < OWFile.SIZE_LIMIT:
            missing_in_attr = missing_values(table.get_nan_frequency_attribute())
            missing_in_class = missing_values(table.get_nan_frequency_class())
        nattrs = len(domain.attributes)
        text += f"<br/>{nattrs} {pl(nattrs, 'feature')} {missing_in_attr}"
        if domain.has_continuous_class:
            text += f"<br/>Regression; numerical class {missing_in_class}"
        elif domain.has_discrete_class:
            nvals = len(domain.class_var.values)
            text += "<br/>Classification; categorical class " \
                f"with {nvals} {pl(nvals, 'value')} {missing_in_class}"
        elif table.domain.class_vars:
            ntargets = len(table.domain.class_vars)
            text += "<br/>Multi-target; " \
                f"{ntargets} target {pl(ntargets, 'variable')} " \
                f"{missing_in_class}"
        else:
            text += "<br/>Data has no target variable."
        nmetas = len(domain.metas)
        text += f"<br/>{nmetas} {pl(nmetas, 'meta attribute')}"
        text += "</p>"

        if 'Timestamp' in table.domain:
            # Google Forms uses this header to timestamp responses
            text += f"<p>First entry: {table[0, 'Timestamp']}<br/>" \
                f"Last entry: {table[-1, 'Timestamp']}</p>"
        return text

    def storeSpecificSettings(self):
        self.current_context.modified_variables = self.variables[:]

    def retrieveSpecificSettings(self):
        if hasattr(self.current_context, "modified_variables"):
            self.variables[:] = self.current_context.modified_variables

    def reset_domain_edit(self):
        self.domain_editor.reset_domain()
        self.apply_domain_edit()

    def _inspect_discrete_variables(self, domain):
        for var in chain(domain.variables, domain.metas):
            if var.is_discrete and len(var.values) > 100:
                self.Warning.performance_warning()

    def apply_domain_edit(self):
        self.Warning.performance_warning.clear()
        self.Warning.renamed_vars.clear()
        if self.data is None:
            table = None
        else:
            domain, cols, renamed = \
                self.domain_editor.get_domain(self.data.domain, self.data,
                                              deduplicate=True)
            if not (domain.variables or domain.metas):
                table = None
            elif domain is self.data.domain:
                table = self.data
            else:
                X, y, m = cols
                table = Table.from_numpy(domain, X, y, m, self.data.W)
                table.name = self.data.name
                table.ids = np.array(self.data.ids)
                table.attributes = getattr(self.data, 'attributes', {})
                self._inspect_discrete_variables(domain)
            if renamed:
                self.Warning.renamed_vars(f"Renamed: {', '.join(renamed)}")

        self.Warning.multiple_targets(
            shown=table is not None and len(table.domain.class_vars) > 1)
        self.Outputs.data.send(table)
        self.apply_button.setEnabled(False)

    def get_widget_name_extension(self):
        _, name = os.path.split(self.loaded_file)
        return os.path.splitext(name)[0]

    def send_report(self):
        def get_ext_name(filename):
            try:
                return FileFormat.names[os.path.splitext(filename)[1]]
            except KeyError:
                return "unknown"

        if self.data is None:
            self.report_paragraph("File", "No file.")
            return

        if self.source == self.LOCAL_FILE:
            home = os.path.expanduser("~")
            if self.loaded_file.startswith(home):
                # os.path.join does not like ~
                name = "~" + os.path.sep + \
                       self.loaded_file[len(home):].lstrip("/").lstrip("\\")
            else:
                name = self.loaded_file
            if self.sheet_combo.isVisible():
                name += f" ({self.sheet_combo.currentText()})"
            self.report_items("File", [("File name", name),
                                       ("Format", get_ext_name(name))])
        else:
            self.report_items("Data", [("Resource", self.url),
                                       ("Format", get_ext_name(self.url))])

        self.report_data("Data", self.data)

    @staticmethod
    def dragEnterEvent(event):
        """Accept drops of valid file urls"""
        urls = event.mimeData().urls()
        if urls:
            try:
                FileFormat.get_reader(urls[0].toLocalFile())
                event.acceptProposedAction()
            except MissingReaderException:
                pass

    def dropEvent(self, event):
        """Handle file drops"""
        urls = event.mimeData().urls()
        if urls:
            self.add_path(urls[0].toLocalFile())  # add first file
            self.source = self.LOCAL_FILE
            self.load_data()

    def workflowEnvChanged(self, key, value, oldvalue):
        """
        Function called when environment changes (e.g. while saving the scheme)
        It make sure that all environment connected values are modified
        (e.g. relative file paths are changed)
        """
        self.update_file_list(key, value, oldvalue)


class OWFileDropHandler(SingleUrlDropHandler):
    WIDGET = OWFile

    def canDropUrl(self, url: QUrl) -> bool:
        if url.isLocalFile():
            try:
                FileFormat.get_reader(url.toLocalFile())
                return True
            except Exception:  # noqa # pylint:disable=broad-except
                return False
        else:
            return url.scheme().lower() in ("http", "https", "ftp")

    def parametersFromUrl(self, url: QUrl) -> Dict[str, Any]:
        if url.isLocalFile():
            path = url.toLocalFile()
            r = RecentPath(os.path.abspath(path), None, None,
                           os.path.basename(path))
            return {
                "recent_paths": stored_recent_paths_prepend(self.WIDGET, r),
                "source": OWFile.LOCAL_FILE,
            }
        else:
            return {
                "recent_urls": [url.toString()],
                "source": OWFile.URL,
            }


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWFile).run()



# pylint: disable=protected-access,duplicate-code
import copy
import pkg_resources
import unittest
from unittest.mock import Mock

import numpy as np
import sklearn

from AnyQt.QtGui import QFont, QPen

from Orange.classification import ThresholdClassifier
from Orange.data import Table
import Orange.evaluation
import Orange.classification

from Orange.widgets.evaluate.tests.base import EvaluateTest
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.tests.utils import simulate
from Orange.widgets.evaluate.owliftcurve import OWLiftCurve, cumulative_gains, \
    cumulative_gains_from_results, CurveTypes, precision_recall_from_results, \
    points_from_results, compute_area
from Orange.tests import test_filename


# scikit-learn==1.1.1 does not support read the docs, therefore
# we can not make it a requirement for now. When the minimum required
# version is >=1.1.1, delete these exceptions.
OK_SKLEARN = pkg_resources.parse_version(sklearn.__version__) >= \
             pkg_resources.parse_version("1.1.1")
SKIP_REASON = "Only test precision-recall with scikit-learn>=1.1.1"


class TestOWLiftCurve(WidgetTest, EvaluateTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.lenses = data = Table(test_filename("datasets/lenses.tab"))
        test_on_test = Orange.evaluation.TestOnTestData(store_data=True)
        cls.res = test_on_test(
            data=data[::2], test_data=data[1::2],
            learners=[Orange.classification.MajorityLearner(),
                      Orange.classification.KNNLearner()]
        )

    def setUp(self):
        super().setUp()
        self.widget = self.create_widget(
            OWLiftCurve,
            stored_settings={
                "display_convex_hull": True
            }
        )  # type: OWLiftCurve

    def test_basic(self):
        self.send_signal(self.widget.Inputs.evaluation_results, self.res)
        simulate.combobox_run_through_all(self.widget.target_cb)

    def test_empty_input(self):
        res = copy.copy(self.res)
        res.actual = res.actual[:0]
        res.row_indices = res.row_indices[:0]
        res.predicted = res.predicted[:, :0]
        res.probabilities = res.probabilities[:, :0, :]
        self.send_signal(self.widget.Inputs.evaluation_results, res)

    def test_nan_input(self):
        res = copy.copy(self.res)
        res.actual = res.actual.copy()
        res.actual[0] = np.nan
        self.send_signal(self.widget.Inputs.evaluation_results, res)
        self.assertTrue(self.widget.Error.invalid_results.is_shown())
        self.send_signal(self.widget.Inputs.evaluation_results, None)
        self.assertFalse(self.widget.Error.invalid_results.is_shown())

    def test_cumulative_gains(self):
        self.send_signal(self.widget.Inputs.evaluation_results, self.res)
        radio_buttons = self.widget.controls.curve_type.buttons
        radio_buttons[CurveTypes.CumulativeGains].click()
        self.assertEqual(self.widget.curve_type, CurveTypes.CumulativeGains)

    def test_precision_recall(self):
        self.send_signal(self.widget.Inputs.evaluation_results, self.res)
        radio_buttons = self.widget.controls.curve_type.buttons
        radio_buttons[CurveTypes.PrecisionRecall].click()
        self.assertEqual(self.widget.curve_type, CurveTypes.PrecisionRecall)

    def test_get_threshold(self):
        recall = np.array([1, 2 / 3, 2 / 3, 1 / 3, 0])
        thresholds = np.array([0.4, 0.5, 0.6, 0.9, 1])

        self.widget.rate = 1
        threshold = self.widget._get_threshold(recall, thresholds)
        self.assertEqual(threshold, 0.4)

        self.widget.rate = 0.7
        threshold = self.widget._get_threshold(recall, thresholds)
        self.assertEqual(threshold, 0.4)

        self.widget.rate = 0.5
        threshold = self.widget._get_threshold(recall, thresholds)
        self.assertEqual(threshold, 0.6)

        self.widget.rate = 0.3
        threshold = self.widget._get_threshold(recall, thresholds)
        self.assertEqual(threshold, 0.9)

        self.widget.rate = 0
        threshold = self.widget._get_threshold(recall, thresholds)
        self.assertEqual(threshold, 1)

    def test_threshold_tooltip(self):
        data = Table("heart_disease")
        test_on_test = Orange.evaluation.TestOnTestData(
            store_data=True, store_models=True)
        res = test_on_test(data=data[::2], test_data=data[1::2],
                           learners=[Orange.classification.MajorityLearner(),
                                     Orange.classification.KNNLearner()])
        self.send_signal(self.widget.Inputs.evaluation_results, res)
        self.assertEqual(self.widget.tooltip.toPlainText(),
                         "Probability threshold(s):\n— 0.526\n— 0.4")

        self.widget.line.setPos(0.9)
        self.assertEqual(self.widget.tooltip.toPlainText(),
                         "Probability threshold(s):\n— 0.526\n— 0.2")

        self.widget.line.setPos(0.0)
        self.assertEqual(self.widget.tooltip.toPlainText(),
                         "Probability threshold(s):\n— 0.526\n— 1.0")

    def test_point_tooltip(self):
        data = Table("heart_disease")
        test_on_test = Orange.evaluation.TestOnTestData(
            store_data=True, store_models=True)
        res = test_on_test(data=data[::2], test_data=data[1::2],
                           learners=[Orange.classification.MajorityLearner(),
                                     Orange.classification.KNNLearner()])
        self.send_signal(self.widget.Inputs.evaluation_results, res)
        scatter = self.widget.plot.curve_items[-1].scatter

        vb = scatter.getViewBox()
        vb.setToolTip = Mock()

        ev = Mock()
        ev.exit = False
        scatter._maskAt = Mock(side_effect=
                               lambda *_: np.array([1] + 5 * [0], dtype=bool))
        scatter.hoverEvent(ev)
        text = 'P Rate: 0.086\nLift: 1.521\nThreshold: 1.0'
        vb.setToolTip.assert_called_with(text)

    def test_output(self):
        data = Table("heart_disease")
        test_on_test = Orange.evaluation.TestOnTestData(
            store_data=True, store_models=True)
        res = test_on_test(data=data[::2], test_data=data[1::2],
                           learners=[Orange.classification.MajorityLearner(),
                                     Orange.classification.KNNLearner()])
        self.send_signal(self.widget.Inputs.evaluation_results, res)
        model = self.get_output(self.widget.Outputs.calibrated_model)
        self.assertIsNone(model)
        self.assertTrue(self.widget.Information.no_output.is_shown())

        self.widget.selected_classifiers = [1]
        self.widget._on_classifiers_changed()
        model = self.get_output(self.widget.Outputs.calibrated_model)
        self.assertIsInstance(model, ThresholdClassifier)
        self.assertEqual(model.threshold, 0.6)
        self.assertFalse(self.widget.Information.no_output.is_shown())

    @WidgetTest.skipNonEnglish
    def test_visual_settings(self):
        graph = self.widget.plot

        def test_settings():
            font = QFont("Helvetica", italic=True, pointSize=20)
            self.assertFontEqual(
                graph.parameter_setter.title_item.item.font(), font
            )

            font.setPointSize(16)
            for item in graph.parameter_setter.axis_items:
                self.assertFontEqual(item.label.font(), font)

            font.setPointSize(15)
            for item in graph.parameter_setter.axis_items:
                self.assertFontEqual(item.style["tickFont"], font)

            self.assertEqual(
                graph.parameter_setter.title_item.item.toPlainText(), "Foo"
            )
            self.assertEqual(graph.parameter_setter.title_item.text, "Foo")

            for line in graph.curve_items:
                pen: QPen = line.opts["pen"]
                self.assertEqual(pen.width(), 10)

            pen: QPen = graph.default_line_item.opts["pen"]
            self.assertEqual(pen.width(), 4)

        test_on_test = Orange.evaluation.TestOnTestData(store_data=True)
        res = test_on_test(
            data=self.lenses[::2], test_data=self.lenses[1::2],
            learners=[Orange.classification.MajorityLearner(),
                      Orange.classification.KNNLearner()]
        )
        self.send_signal(self.widget.Inputs.evaluation_results, res)
        key, value = ("Fonts", "Font family", "Font family"), "Helvetica"
        self.widget.set_visual_settings(key, value)

        key, value = ("Fonts", "Title", "Font size"), 20
        self.widget.set_visual_settings(key, value)
        key, value = ("Fonts", "Title", "Italic"), True
        self.widget.set_visual_settings(key, value)

        key, value = ("Fonts", "Axis title", "Font size"), 16
        self.widget.set_visual_settings(key, value)
        key, value = ("Fonts", "Axis title", "Italic"), True
        self.widget.set_visual_settings(key, value)

        key, value = ("Fonts", "Axis ticks", "Font size"), 15
        self.widget.set_visual_settings(key, value)
        key, value = ("Fonts", "Axis ticks", "Italic"), True
        self.widget.set_visual_settings(key, value)

        key, value = ("Annotations", "Title", "Title"), "Foo"
        self.widget.set_visual_settings(key, value)

        key, value = ("Figure", "Line", "Width"), 10
        self.widget.set_visual_settings(key, value)

        key, value = ("Figure", "Default Line", "Width"), 4
        self.widget.set_visual_settings(key, value)

        self.send_signal(self.widget.Inputs.evaluation_results, res)
        test_settings()

        self.send_signal(self.widget.Inputs.evaluation_results, None)
        self.send_signal(self.widget.Inputs.evaluation_results, res)
        test_settings()

    def assertFontEqual(self, font1, font2):
        self.assertEqual(font1.family(), font2.family())
        self.assertEqual(font1.pointSize(), font2.pointSize())
        self.assertEqual(font1.italic(), font2.italic())


class UtilsTest(unittest.TestCase):
    @staticmethod
    def test_cumulative_gains():
        shuffle = [1, 2, 0, 3, 5, 4]
        y_true = np.array([1, 1, 0, 0, 1, 0])[shuffle]
        y_scores = np.array([0.9, 0.6, 0.5, 0.4, 0.4, 0.2])[shuffle]

        assert_almost_equal = np.testing.assert_almost_equal

        contacted, respondents, thresholds = cumulative_gains(y_true, y_scores)
        assert_almost_equal(contacted, np.array([1, 2, 3, 5, 6]) / 6)
        assert_almost_equal(thresholds, [0.9, 0.6, 0.5, 0.4, 0.2])
        assert_almost_equal(respondents, np.array([1, 2, 2, 3, 3]) / 3)

        contacted, respondents, thresholds = cumulative_gains(y_true, 1 - y_scores, target=0)
        assert_almost_equal(contacted, np.array([1, 3, 4, 5, 6]) / 6)
        assert_almost_equal(thresholds, [0.8, 0.6, 0.5, 0.4, 0.1])
        assert_almost_equal(respondents, np.array([1, 2, 3, 3, 3]) / 3)

        contacted, respondents, thresholds = \
            cumulative_gains(np.array([], dtype=int), np.array([]))
        assert_almost_equal(contacted, [])
        assert_almost_equal(respondents, [])
        assert_almost_equal(thresholds, [])

    @staticmethod
    def test_cumulative_gains_from_results():
        shuffle = [1, 2, 0, 3, 5, 4]
        y_true = np.array([1, 1, 0, 0, 1, 0])[shuffle]
        y_scores = np.array([0.9, 0.6, 0.5, 0.4, 0.4, 0.2])[shuffle]

        results = Mock()
        results.actual = y_true
        results.probabilities = \
            [Mock(), Mock(), np.vstack((1 - y_scores, y_scores)).T]

        assert_almost_equal = np.testing.assert_almost_equal

        contacted, respondents, thresholds = \
            cumulative_gains_from_results(results, 1, 2)
        assert_almost_equal(thresholds, [0.9, 0.6, 0.5, 0.4, 0.2])
        assert_almost_equal(contacted, np.array([1, 2, 3, 5, 6]) / 6)
        assert_almost_equal(respondents, np.array([1, 2, 2, 3, 3]) / 3)

        contacted, respondents, thresholds = \
            cumulative_gains_from_results(results, 0, 2)
        assert_almost_equal(contacted, np.array([1, 3, 4, 5, 6]) / 6)
        assert_almost_equal(thresholds, [0.8, 0.6, 0.5, 0.4, 0.1])
        assert_almost_equal(respondents, np.array([1, 2, 3, 3, 3]) / 3)

        results.actual = np.array([], dtype=int)
        results.probabilities = np.empty((3, 0, 2))
        contacted, respondents, thresholds = \
            cumulative_gains(np.array([], dtype=int), np.array([]))
        assert_almost_equal(contacted, [])
        assert_almost_equal(respondents, [])
        assert_almost_equal(thresholds, [])

    @staticmethod
    @unittest.skipUnless(OK_SKLEARN, SKIP_REASON)
    def test_precision_recall_from_results():
        y_true = np.array([1, 0, 1, 0, 0, 1])
        y_scores = np.array([0.6, 0.5, 0.9, 0.4, 0.2, 0.4])

        results = Mock()
        results.actual = y_true
        results.probabilities = \
            [Mock(), Mock(), np.vstack((1 - y_scores, y_scores)).T]

        recall, precision, thresholds = \
            precision_recall_from_results(results, 1, 2)
        np.testing.assert_equal(precision,
                                np.array([1 / 2, 3 / 5, 2 / 3, 1, 1, 1]))
        np.testing.assert_equal(recall,
                                np.array([1, 1, 2 / 3, 2 / 3, 1 / 3, 0]))
        np.testing.assert_equal(thresholds,
                                np.array([0.2, 0.4, 0.5, 0.6, 0.9, 1]))

    @staticmethod
    @unittest.skipUnless(OK_SKLEARN, SKIP_REASON)
    def test_precision_recall_from_results_one():
        y_true = np.array([1, 0, 1, 0, 0, 1])
        y_scores = np.array([0.6, 0.5, 1, 0.4, 0.2, 0.4])

        results = Mock()
        results.actual = y_true
        results.probabilities = \
            [Mock(), Mock(), np.vstack((1 - y_scores, y_scores)).T]

        recall, precision, thresholds = \
            precision_recall_from_results(results, 1, 2)
        np.testing.assert_equal(precision,
                                np.array([1 / 2, 3 / 5, 2 / 3, 1, 1]))
        np.testing.assert_equal(recall,
                                np.array([1, 1, 2 / 3, 2 / 3, 1 / 3]))
        np.testing.assert_equal(thresholds,
                                np.array([0.2, 0.4, 0.5, 0.6, 1]))

    @staticmethod
    @unittest.skipUnless(OK_SKLEARN, SKIP_REASON)
    def test_precision_recall_from_results_multiclass():
        y_true = np.array([1, 0, 1, 0, 2, 2])
        y_scores = np.array([[0.3, 0.3, 0.4],
                             [0.3, 0.4, 0.4],
                             [0.1, 0.9, 0.1],
                             [0.4, 0.2, 0.4],
                             [0.1, 0.2, 0.7],
                             [0.1, 0.1, 0.8]])

        results = Mock()
        results.actual = y_true
        results.probabilities = [Mock(), Mock(), y_scores]

        recall, precision, thresholds = \
            precision_recall_from_results(results, 1, 2)
        np.testing.assert_equal(precision,
                                np.array([1 / 3, 2 / 5, 2 / 3, 1 / 2, 1, 1]))
        np.testing.assert_equal(recall, np.array([1, 1, 1, 1 / 2, 1 / 2, 0]))
        np.testing.assert_equal(thresholds,
                                np.array([0.1, 0.2, 0.3, 0.4, 0.9, 1]))

    @staticmethod
    def test_points_from_results_cumulative_gain():
        y_scores = np.array([0.6, 0.5, 0.9, 0.4, 0.2, 0.4])
        results = Mock()
        results.actual = np.array([1, 0, 1, 0, 0, 1])
        results.probabilities = \
            [Mock(), Mock(), np.vstack((1 - y_scores, y_scores)).T]

        contacted, respondents, thresholds = \
            cumulative_gains_from_results(results, 1, 2)
        res = points_from_results(results, 1, 2, CurveTypes.CumulativeGains)
        np.testing.assert_almost_equal(res.contacted, contacted)
        np.testing.assert_almost_equal(res.respondents, respondents)
        np.testing.assert_almost_equal(res.thresholds, thresholds)

    @staticmethod
    def test_points_from_results_precision_recall():
        y_scores = np.array([0.6, 0.5, 0.9, 0.4, 0.2, 0.4])
        results = Mock()
        results.actual = np.array([1, 0, 1, 0, 0, 1])
        results.probabilities = \
            [Mock(), Mock(), np.vstack((1 - y_scores, y_scores)).T]

        contacted, respondents, thresholds = \
            precision_recall_from_results(results, 1, 2)
        res = points_from_results(results, 1, 2, CurveTypes.PrecisionRecall)
        np.testing.assert_almost_equal(res.contacted, contacted)
        np.testing.assert_almost_equal(res.respondents, respondents)
        np.testing.assert_almost_equal(res.thresholds, thresholds)

    def test_area(self):
        x = np.array([5, 8, 9, 11])
        y = np.array([7, 14, 8, 0])
        area = compute_area(x, y)
        self.assertEqual(area, 51)


if __name__ == "__main__":
    unittest.main()



import warnings

from sklearn.exceptions import ConvergenceWarning

from Orange.widgets.model.owneuralnetwork import OWNNLearner
from Orange.widgets.tests.base import WidgetTest, WidgetLearnerTestMixin


class TestOWNeuralNetwork(WidgetTest, WidgetLearnerTestMixin):
    def setUp(self):
        warnings.filterwarnings("ignore", ".*", ConvergenceWarning)
        self.widget = self.create_widget(OWNNLearner,
                                         stored_settings={"auto_apply": False})
        self.init()

    def test_migrate_setting(self):
        settings = dict(alpha=2.9)
        OWNNLearner.migrate_settings(settings, 0)
        self.assertEqual(OWNNLearner.alphas[settings["alpha_index"]], 3)

        settings = dict(alpha=103)
        OWNNLearner.migrate_settings(settings, 0)
        self.assertEqual(OWNNLearner.alphas[settings["alpha_index"]], 100)

        settings = dict(alpha_index=0)
        OWNNLearner.migrate_settings(settings, version=1)
        self.assertEqual(OWNNLearner.alphas[settings["alpha_index"]], 0.0001)

    def test_no_layer_warning(self):
        self.assertFalse(self.widget.Warning.no_layers.is_shown())

        self.widget.hidden_layers_input = ""
        self.click_apply()
        self.assertTrue(self.widget.Warning.no_layers.is_shown())

        self.widget.hidden_layers_input = "10,"
        self.click_apply()
        self.assertFalse(self.widget.Warning.no_layers.is_shown())



import os
import unittest

import numpy as np

from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.unsupervised.owdistancefile \
    import OWDistanceFile, OWDistanceFileDropHandler

import Orange.tests


class TestOWDistanceFile(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWDistanceFile)

    def open_file(self, filename):
        filename = os.path.join(os.path.split(Orange.tests.__file__)[0],
                                filename)
        self.widget.add_path(filename)
        self.widget.open_file()

    def test_non_square(self):
        self.open_file("xlsx_files/distances_nonsquare.xlsx")
        self.assertIsNone(self.get_output(self.widget.Outputs.distances))
        self.assertTrue(self.widget.Error.non_square_matrix.is_shown())
        self.open_file("xlsx_files/distances_with_nans.xlsx")
        self.assertFalse(self.widget.Error.non_square_matrix.is_shown())

    def test_nan_to_num(self):
        self.open_file("xlsx_files/distances_with_nans.xlsx")
        dist = self.get_output(self.widget.Outputs.distances)
        np.testing.assert_equal(dist, [[1, 2, 3], [4, 5, 0], [7, 0, 9]])


class TestOWDistanceFileDropHandler(unittest.TestCase):
    def test_canDropFile(self):
        handler = OWDistanceFileDropHandler()
        self.assertTrue(handler.canDropFile("test.dst"))
        self.assertTrue(handler.canDropFile("test.xlsx"))
        self.assertFalse(handler.canDropFile("test.bin"))

    def test_parametersFromFile(self):
        handler = OWDistanceFileDropHandler()
        r = handler.parametersFromFile("test.dst")
        self.assertEqual(r["recent_paths"][0].basename, "test.dst")



"""
Utility widgets/helper for reading csv like files.

Contents
--------

* CSVOptionsWidget
  Edit options for interpreting a csv file

* CSVImportWidget
  Read and preview part of the file

* TablePreviewModel
  An QAbstractTableModel feeding data from a csv.reader like rows iterator
  implementing lazy iterative loading (`QAbstractItemModel.fetchMore`)

"""
# TODO: Consider a wizard-like interface:
#   * 1. Select encoding, delimiter, ... (preview is all text)
#   * 2. Define column types (preview is parsed and rendered type appropriate)

import sys
import io
import enum
import codecs
import csv
import traceback
import itertools

from functools import singledispatch
from collections import defaultdict
from types import MappingProxyType

import typing
from typing import (
    List, Tuple, Dict, Iterator, Optional, Any, Union, Callable, Mapping
)

from AnyQt.QtCore import (
    Qt, QSize, QPoint, QRect, QRectF, QRegularExpression, QAbstractTableModel,
    QModelIndex, QItemSelectionModel, QTextBoundaryFinder, QTimer, QEvent
)
from AnyQt.QtCore import pyqtSignal as Signal, pyqtSlot as Slot
from AnyQt.QtGui import (
    QRegularExpressionValidator, QColor, QBrush, QPalette, QHelpEvent,
    QStandardItemModel, QStandardItem, QIcon, QIconEngine, QPainter, QPixmap,
    QFont
)
from AnyQt.QtWidgets import (
    QWidget, QComboBox, QFormLayout, QHBoxLayout, QVBoxLayout, QLineEdit,
    QHeaderView, QFrame, QTableView, QMenu, QLabel, QAction, QActionGroup,
    QStyleOptionFrame, QStyle, QStyledItemDelegate, QStyleOptionViewItem,
    QApplication, QAbstractItemView, QToolTip, QStyleOption
)

from Orange.widgets.utils import encodings
from Orange.widgets.utils.overlay import OverlayWidget
from Orange.widgets.utils.combobox import TextEditCombo


__all__ = ["ColumnType", "RowSpec", "CSVOptionsWidget", "CSVImportWidget"]

if typing.TYPE_CHECKING:
    # pylint: disable=invalid-name
    _A = typing.TypeVar("_A")
    _B = typing.TypeVar("_B")


class StampIconEngine(QIconEngine):
    def __init__(self, char, brush):
        # type: (str, Union[QBrush, QColor]) -> None
        super().__init__()
        self.__char = char
        self.__brush = QBrush(brush)

    def paint(self, painter, rect, mode, state):
        # type: (QPainter, QRect, QIcon.Mode, QIcon.State) -> None
        size = rect.size()
        if size.isNull():
            return  # pragma: no cover
        dpr = 1.0
        try:
            dpr = painter.device().devicePixelRatioF()
        except AttributeError:
            pass
        if dpr != 1.0:
            size = size * dpr
        painter.drawPixmap(rect, self.pixmap(size, mode, state))

    def pixmap(self, size, mode, state):
        # type: (QSize, QIcon.Mode, QIcon.State) -> QPixmap
        pm = QPixmap(size)
        pm.fill(Qt.transparent)
        painter = QPainter(pm)
        painter.setRenderHints(
            QPainter.Antialiasing | QPainter.TextAntialiasing |
            QPainter.SmoothPixmapTransform
        )
        size = size.width()
        color = self.__brush.color()
        painter.setPen(color)
        painter.setBrush(color)
        margin = 1 + size // 16
        text_margin = size // 20
        rect = QRectF(margin, margin, size - 2 * margin, size - 2 * margin)
        painter.drawRoundedRect(rect, 30.0, 30.0, Qt.RelativeSize)
        painter.setPen(Qt.white)

        font = painter.font()  # type: QFont
        font.setPixelSize(size - 2 * margin - 2 * text_margin)
        font.setBold(True)
        painter.setFont(font)

        painter.drawText(rect, Qt.AlignCenter, self.__char)
        painter.end()

        style = QApplication.style()
        if style is not None:
            opt = QStyleOption()
            opt.palette = QApplication.palette()
            pm = style.generatedIconPixmap(mode, pm, opt)
        return pm


class Dialect(csv.Dialect):
    def __init__(self, delimiter, quotechar, escapechar, doublequote,
                 skipinitialspace, quoting=csv.QUOTE_MINIMAL):
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.escapechar = escapechar
        self.doublequote = doublequote
        self.skipinitialspace = skipinitialspace
        self.quoting = quoting
        self.lineterminator = "\r\n"
        super().__init__()

    def __repr__(self):
        _, args, *_ = self.__reduce__()
        args = ", ".join(map("{!r}".format, args))
        return "Dialect(" + args + ")"

    def __reduce__(self):
        return type(self), (self.delimiter, self.quotechar, self.escapechar,
                            self.doublequote, self.skipinitialspace,
                            self.quoting)


class ColumnType(enum.Enum):
    """
    Possible column types
    """
    # Skip column
    Skip = "Skip"
    # Autodetect column type
    Auto = "Auto"
    # Numeric (real) column
    Numeric = "Numeric"
    # Categorical column
    Categorical = "Categorical"
    # Text column
    Text = "Text"
    # Date time column
    Time = "Time"


class LineEdit(QLineEdit):
    """
    A line edit widget with a `minimumContentsLength` property.

    Similar to QComboBox.minimumContentsLength
    """
    # These constants are taken from Qt's sources for QLineEdit
    _verticalMargin = 1
    _horizontalMargin = 2

    def __init__(self, *args, minimumContentsLength=0, **kwargs):
        self.__minimumContentsLength = minimumContentsLength
        super().__init__(*args, **kwargs)

    def setMinimumContentsLength(self, characters):
        # type: (int) -> None
        """
        Set the minimum number of characters that should fit into the
        line edit (used for size hinting).
        """
        if self.__minimumContentsLength != characters:
            self.__minimumContentsLength = characters
            self.updateGeometry()

    def minimumContentsLength(self):
        # type: () -> int
        """
        Return the minimum number of characters that should fit into the
        line edit.
        """
        return self.__minimumContentsLength

    def sizeHint(self):
        # type: () -> QSize
        """Reimplemented."""
        # Most of this is taken from QLineEdit's sources, the difference
        # is only in the use of minimumContentsLength instead of a
        # hardcoded constant.
        self.ensurePolished()
        fm = self.fontMetrics()
        textmargins = self.textMargins()
        contentsmargins = self.contentsMargins()

        h = (max(fm.height(), 14) + 2 * self._verticalMargin +
             textmargins.top() + textmargins.bottom() +
             contentsmargins.top() + contentsmargins.bottom())

        nchar = self.__minimumContentsLength
        if nchar <= 0:
            nchar = 17

        w = (fm.horizontalAdvance("X") * nchar + 2 * self._horizontalMargin +
             textmargins.left() + textmargins.right() +
             contentsmargins.left() + contentsmargins.right())

        opt = QStyleOptionFrame()
        self.initStyleOption(opt)
        size = self.style().sizeFromContents(
            QStyle.CT_LineEdit, opt, QSize(w, h), self
        )
        return size

    def minimumSizeHint(self):
        # type: () -> QSize
        """Reimplemented."""
        if self.__minimumContentsLength > 0:
            return self.sizeHint()
        else:
            return super(LineEdit, self).sizeHint()


class CSVOptionsWidget(QWidget):
    """
    A widget presenting common CSV options.
    """
    DelimiterTab, DelimiterComma, DelimiterSemicolon, DelimiterSpace = range(4)
    DelimiterOther = DelimiterSpace + 2  # note DelimiterSpace + 1 is reserved

    PresetDelimiters = [
        ("Tab", "\t"),
        ("Comma", ","),
        ("Semicolon", ";"),
        ("Space", " "),
    ]

    #: Signal emitted when the format (dialect) changes
    optionsChanged = Signal()
    #: Signal emitted when the format (dialect) is edited by the user
    optionsEdited = Signal()

    def __init__(self, *args, **kwargs):
        self._delimiter_idx = 0
        self._delimiter = ","
        self._delimiter_custom = "|"
        self._quotechar = "\""
        self._encoding = kwargs.pop("selectedEncoding", "utf-8")

        super().__init__(*args, **kwargs)

        # Dialect options form
        form = QFormLayout()
        self.encoding_cb = QComboBox(
            objectName="encoding-combo-box",
            toolTip="Select file text encoding",
        )
        self.__set_visible_codecs(encodings.list_selected_encodings())
        self.encoding_cb.activated.connect(self.__on_encoding_activated)

        self.delimiter_cb = QComboBox(
            objectName="delimiter-combo-box",
            toolTip="Select cell delimiter character."
        )
        self.delimiter_cb.addItems(
            [name for name, _ in CSVOptionsWidget.PresetDelimiters]
        )
        self.delimiter_cb.insertSeparator(self.delimiter_cb.count())
        self.delimiter_cb.addItem("Other")

        self.delimiter_cb.setCurrentIndex(self._delimiter_idx)
        self.delimiter_cb.activated.connect(self.__on_delimiter_idx_activated)

        validator = QRegularExpressionValidator(QRegularExpression("."))
        self.delimiteredit = LineEdit(
            self._delimiter_custom,
            enabled=self._delimiter_idx == CSVOptionsWidget.DelimiterOther,
            minimumContentsLength=2,
            objectName="custom-delimiter-edit"
        )
        self.delimiteredit.setValidator(validator)
        self.delimiteredit.editingFinished.connect(self.__on_delimiter_edited)

        delimlayout = QHBoxLayout()
        delimlayout.setContentsMargins(0, 0, 0, 0)
        delimlayout.addWidget(self.delimiter_cb)
        delimlayout.addWidget(self.delimiteredit)
        self.quoteedit = TextEditCombo(
            editable=True, minimumContentsLength=1,
            sizeAdjustPolicy=QComboBox.AdjustToMinimumContentsLengthWithIcon,
            objectName="quote-edit-combo-box"
        )
        self.quoteedit.addItems(["\"", "'"])
        self.quoteedit.setValidator(validator)
        self.quoteedit.setText(self._quotechar)
        self.quoteedit.activated.connect(self.__on_quotechar_edited)

        quotelayout = QHBoxLayout()
        quotelayout.setContentsMargins(0, 0, 0, 0)
        quotelayout.addWidget(self.quoteedit)

        form.addRow("Encoding", self.encoding_cb)
        form.addRow(QFrame(self, frameShape=QFrame.HLine))
        form.addRow("Cell delimiter", delimlayout)
        form.addRow("Quote character", self.quoteedit)
        self.setLayout(form)

    def dialect(self):
        # type: () -> csv.Dialect
        """
        Return the current state as a `csv.Dialect` instance.
        """
        if self._delimiter_idx == CSVOptionsWidget.DelimiterOther:
            delimiter = self._delimiter_custom
        elif 0 <= self._delimiter_idx < len(CSVOptionsWidget.PresetDelimiters):
            _, delimiter = CSVOptionsWidget.PresetDelimiters[self._delimiter_idx]
        else:
            assert False

        quotechar = self.quoteedit.text() or None
        skipinitialspace = True
        escapechar = None
        quoting = csv.QUOTE_MINIMAL if quotechar is not None else csv.QUOTE_NONE
        return Dialect(delimiter, quotechar, escapechar,
                       doublequote=True, skipinitialspace=skipinitialspace,
                       quoting=quoting)

    def setDialect(self, dialect):
        # type: (csv.Dialect) -> None
        """
        Set the current state to match dialect instance.
        """
        changed = False
        delimiter = dialect.delimiter
        presets = [d for _, d in CSVOptionsWidget.PresetDelimiters]
        try:
            index = presets.index(delimiter)
        except ValueError:
            index = CSVOptionsWidget.DelimiterOther
            if self._delimiter_custom != delimiter:
                self._delimiter_custom = delimiter
                changed = True

        if self._delimiter_idx != index:
            self._delimiter_idx = index
            self.delimiter_cb.setCurrentIndex(index)
            self.delimiteredit.setText(delimiter)
            changed = True
        if self._quotechar != dialect.quotechar:
            self._quotechar = dialect.quotechar
            self.quoteedit.setText(dialect.quotechar or '')
            changed = True

        if changed:
            self.optionsChanged.emit()

    def setSelectedEncoding(self, encoding):
        # type: (str) -> None
        """
        Set the current selected encoding.

        Parameters
        ----------
        encoding : str
            Encoding name such that `codecs.lookup` finds it.
        """
        co = codecs.lookup(encoding)
        cb = self.encoding_cb
        index = cb.findData(co.name, Qt.UserRole)

        if index == -1:
            # insert the encoding before the separator.
            sepidx = cb.findData("separator", Qt.AccessibleDescriptionRole)
            if sepidx == -1:
                sepidx = cb.count()
            cb.insertItem(sepidx, encodings.display_name(co.name),
                          userData=co.name)
            index = sepidx
            assert cb.itemData(index, Qt.UserRole) == co.name
            changed = True
            self._encoding = encoding
        else:
            changed = index != self.encoding_cb.currentIndex()
            self._encoding = encoding

        self.encoding_cb.setCurrentIndex(index)

        if changed:
            self.optionsChanged.emit()

    def encoding(self):
        # type: () -> str
        """
        Return the current selected encoding.
        """
        index = self.encoding_cb.currentIndex()
        if index >= 0:
            data = self.encoding_cb.itemData(index, Qt.UserRole)
            if isinstance(data, str):
                return data
        return "latin-1"

    def __on_encoding_activated(self, idx):
        current = self._encoding
        data = self.encoding_cb.itemData(idx, Qt.UserRole)
        if data is ...:
            # restore the previous item
            idx = self.encoding_cb.findData(current, Qt.UserRole)
            self.encoding_cb.setCurrentIndex(idx)
            self.__show_encodings_widget()
        elif isinstance(data, str):
            assert codecs.lookup(data)
            self._encoding = data
            self.optionsEdited.emit()
            self.optionsChanged.emit()

    def __show_encodings_widget(self):
        """
        Show the encodings widget for selection
        """
        # If tool window is already shown just raise it
        w = self.findChild(
            encodings.SelectEncodingsWidget,
            "-encoding-selection-tool-window"
        )
        if w is not None and w.isVisible():  # pragma: no coverage
            w.raise_()
            w.activateWindow()
            return

        w = encodings.SelectEncodingsWidget(
            self, Qt.Tool,
            windowTitle="Customize Encodings List",
            objectName="-encoding-selection-tool-window"
        )
        w.setAttribute(Qt.WA_DeleteOnClose)
        model = w.model()
        model.dataChanged.connect(
            lambda: self.__set_visible_codecs(w.selectedEncodings())
        )
        w.show()

    def __set_visible_codecs(self, codecs):
        # type: (List[str]) -> None
        # Set the list of current visible/selectable codecs in the encoding_cb
        if not codecs:
            # never clear all items from the drop down
            codecs = ["ascii"]
        cb = self.encoding_cb
        current = self._encoding
        cb.clear()

        for c in codecs:
            cb.addItem(encodings.display_name(c), userData=c)
        cb.insertSeparator(cb.count())
        cb.addItem("Customize Encodings List...", userData=...)
        idx = cb.findData(current, Qt.UserRole)
        if idx != -1:
            cb.setCurrentIndex(idx)
        elif codecs:
            cb.setCurrentIndex(0)
            self._encoding = codecs[0]
            self.__on_encoding_activated(0)
        else:
            cb.setCurrentIndex(-1)
            self._encoding = ""

    def __on_delimiter_idx_activated(self, index):
        presets = CSVOptionsWidget.PresetDelimiters
        if 0 <= index < CSVOptionsWidget.DelimiterOther:
            self.delimiteredit.setText(presets[index][1])
            self.delimiteredit.setEnabled(False)
        else:
            self.delimiteredit.setText(self._delimiter_custom)
            self.delimiteredit.setEnabled(True)

        if self._delimiter_idx != index:
            self._delimiter_idx = index
            self.optionsChanged.emit()
            self.optionsEdited.emit()

    def __on_delimiter_edited(self):
        delimiter = self.delimiteredit.text()
        if self._delimiter_custom != delimiter:
            self._delimiter_custom = delimiter
            self.optionsChanged.emit()
            self.optionsEdited.emit()

    def __on_quotechar_edited(self):
        quotechar = self.quoteedit.text()
        if self._quotechar != quotechar:
            self._quotechar = quotechar
            self.optionsChanged.emit()
            self.optionsEdited.emit()


class Item(QStandardItem):
    """
    A QStandardItem subclass using a python dict as a backing store.

    Note
    ----
    Unlike QStandardItem, this class does not map `Qt.DisplayRole` and
    `Qt.EditRole` to the same value. Also, accessing or setting via
    `model.itemData` `model.setItemData` and will not work.
    """
    def __init__(self, data=MappingProxyType({})):
        # type: (Mapping[Qt.ItemDataRole, Any]) -> None
        super().__init__()
        self.__data = dict(data)

    def clone(self):
        # type: () -> Item
        """Reimplemented from QStandardItem"""
        return Item(self.__data)

    def setData(self, value, role=Qt.UserRole+1):
        # type: (Any, Qt.ItemDataRole) -> None
        """Reimplemented from QStandardItem"""
        self.__data[role] = value
        model = self.model()
        if model is not None:
            midx = model.indexFromItem(self)
            model.dataChanged.emit(midx, midx, [role])

    def data(self, role=Qt.UserRole+1):
        # type: (Qt.ItemDataRole) -> Any
        """Reimplemented from QStandardItem"""
        if role == Qt.EditRole and role not in self.__data:
            role = Qt.DisplayRole
        elif role == Qt.DisplayRole and role not in self.__data:
            role = Qt.EditRole
        return self.__data.get(role, None)

    def setItemData(self, data):
        # type: (Dict[Qt.ItemDataRole, Any]) -> bool
        roles = list(data.keys())
        self.__data.update(data)
        m = self.model()  #
        if m is not None:
            midx = m.indexFromItem(self)
            m.dataChanged.emit(midx, midx, roles)
        return True

    def itemData(self):
        return self.__data.copy()


class CSVImportWidget(QWidget):
    """
    CSV import widget with a live table preview
    """
    #: Signal emitted on any format option change.
    optionsChanged = Signal()
    #: Signal emitted when a user changes format options.
    optionsEdited = Signal()
    #: Signal emitted when a user changes type affiliation for a column
    columnTypesChanged = Signal()
    #: Signal emitted when the preview content parsing ends with an error.
    #: Note: this does not include errors in cell content interpretation.
    previewReadErrorOccurred = Signal(str)
    #: Signal emitted when the preview model is reset. This is either because
    #: of `setPreviewContents` or a options change.
    previewModelReset = Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.__previewmodel = None  # type: Optional[TablePreviewModel]
        self.__textwrapper = None  # type: Optional[io.TextIOWrapper]
        self.__sample = None
        self.__buffer = None

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        self.optionswidget = CSVOptionsWidget()
        self.optionswidget.optionsChanged.connect(self.optionsChanged)
        self.optionswidget.optionsEdited.connect(self.optionsEdited)

        self.dataview = TablePreview(
            selectionBehavior=QTableView.SelectColumns,
            tabKeyNavigation=False,
        )
        self.dataview.setContextMenuPolicy(Qt.CustomContextMenu)
        self.dataview.customContextMenuRequested.connect(
            self.__dataview_context_menu
        )
        header = self.dataview.horizontalHeader()  # type: QHeaderView
        header.setContextMenuPolicy(Qt.CustomContextMenu)
        header.customContextMenuRequested.connect(
            self.__hheader_context_menu
        )

        header = self.dataview.verticalHeader()
        header.setContextMenuPolicy(Qt.CustomContextMenu)
        header.customContextMenuRequested.connect(
            self.__vheader_context_menu
        )
        style = self.style()
        opt = self.dataview.viewOptions()
        opt.text = "X"
        opt.features |= QStyleOptionViewItem.HasDisplay
        csize = style.sizeFromContents(
            QStyle.CT_ItemViewItem, opt, QSize(18, 18), self.dataview
        )
        header.ensurePolished()
        header.setDefaultSectionSize(max(csize.height(),
                                         header.minimumSectionSize()))
        layout.addWidget(self.optionswidget)
        form = self.optionswidget.layout()
        assert isinstance(form, QFormLayout)
        number_sep_layout = QHBoxLayout()
        self.grouping_sep_edit_cb = TextEditCombo(
            editable=True, objectName="grouping-separator-combo-box",
            toolTip="Thousands group separator",
            minimumContentsLength=1,
            sizeAdjustPolicy=QComboBox.AdjustToMinimumContentsLengthWithIcon
        )
        items = [
            {Qt.DisplayRole: "None", Qt.EditRole: "",
             Qt.ToolTipRole: "No separator"},
            {Qt.DisplayRole: ".", Qt.EditRole: "."},
            {Qt.DisplayRole: ",", Qt.EditRole: ","},
            {Qt.DisplayRole: "Space", Qt.EditRole: " "},
            {Qt.DisplayRole: "'", Qt.EditRole: "'"},
        ]
        m = QStandardItemModel(self)
        m.invisibleRootItem().appendRows([Item(data) for data in items])
        self.grouping_sep_edit_cb.setModel(m)
        # TODO: Treat all space (THIN SPACE, NO-BREAK SPACE, ...) the same?
        # For instance Apple's Numbers exports csv with \N{NO-BREAK SPACE}.
        # Maybe just use unicodedata.normalize('NFKC', ...) as a converter?
        # For now only allow a limited set
        self.grouping_sep_edit_cb.setValidator(
            QRegularExpressionValidator(QRegularExpression(r"(\.|,| |')?"), self)
        )
        self.grouping_sep_edit_cb.textActivated.connect(
            self.__group_sep_activated)

        self.decimal_sep_edit_cb = TextEditCombo(
            editable=True, objectName="decimal-separator-combo-box",
            toolTip="Decimal separator",
            minimumContentsLength=1,
            sizeAdjustPolicy=QComboBox.AdjustToMinimumContentsLengthWithIcon
        )
        self.decimal_sep_edit_cb.setValidator(
            QRegularExpressionValidator(QRegularExpression(r"(\.|,)"), self))
        self.decimal_sep_edit_cb.addItems([".", ","])
        self.decimal_sep_edit_cb.textActivated.connect(
            self.__decimal_sep_activated)

        number_sep_layout.addWidget(QLabel("Grouping:"))
        number_sep_layout.addWidget(self.grouping_sep_edit_cb)
        number_sep_layout.addWidget(QLabel("Decimal:"))
        number_sep_layout.addWidget(self.decimal_sep_edit_cb)
        number_sep_layout.addStretch(10)
        form.addRow("Number separators:", number_sep_layout)
        self.column_type_edit_cb = QComboBox(
            enabled=False, objectName="column-type-edit-combo-box"
        )
        self.column_type_edit_cb.activated.connect(
            self.__on_column_type_edit_activated
        )
        types = [
            {Qt.DisplayRole: "Auto",
             Qt.ToolTipRole: "The type will be determined automatically based "
                             "on column contents.",
             Qt.UserRole: ColumnType.Auto},
            {Qt.DisplayRole: "Numeric", Qt.UserRole: ColumnType.Numeric},
            {Qt.DisplayRole: "Categorical",
             Qt.UserRole: ColumnType.Categorical},
            {Qt.DisplayRole: "Text", Qt.UserRole: ColumnType.Text},
            {Qt.DisplayRole: "Datetime", Qt.UserRole: ColumnType.Time},
            {Qt.AccessibleDescriptionRole: "separator"},
            {Qt.DisplayRole: "Ignore",
             Qt.UserRole: ColumnType.Skip,
             Qt.ToolTipRole: "The column will not be loaded"}
        ]
        typemodel = QStandardItemModel(self)
        for itemdata in types:
            item = Item(itemdata)
            if itemdata.get(Qt.AccessibleDescriptionRole) == "separator":
                item.setFlags(Qt.NoItemFlags)
            typemodel.appendRow(item)

        self.column_type_edit_cb.setModel(typemodel)
        self.column_type_edit_cb.setCurrentIndex(-1)

        form.addRow(QFrame(frameShape=QFrame.HLine))
        form.addRow("Column type", self.column_type_edit_cb)
        layout.addWidget(self.dataview)
        # Overlay error message widget in the bottom left corner of the data
        # view
        self.__overlay = overlay = OverlayWidget(
            parent=self.dataview.viewport(),
            alignment=Qt.AlignBottom | Qt.AlignLeft,
            objectName="-error-overlay",
            visible=False,
        )
        overlay.setLayout(QVBoxLayout())
        self.__error_label = label = QLabel(objectName="-error-text-label")
        overlay.layout().addWidget(label)
        overlay.setWidget(self.dataview.viewport())

        self.setLayout(layout)

        self.__timer = QTimer(self, singleShot=True)
        self.__timer.timeout.connect(self.__resetPreview)
        self.optionswidget.optionsChanged.connect(self.__timer.start)

    def setDialect(self, dialect):
        # type: (csv.Dialect) -> None
        """
        Set the current state to match dialect instance.
        """
        self.optionswidget.setDialect(dialect)

    def dialect(self):
        # type: () -> csv.Dialect
        """
        Return the current dialect.
        """
        return self.optionswidget.dialect()

    def setEncoding(self, encoding):
        # type: (str) -> None
        """Set the current text encoding."""
        self.optionswidget.setSelectedEncoding(encoding)

    def encoding(self):
        # type: () -> str
        """Return the curent text encoding."""
        return self.optionswidget.encoding()

    def setNumbersFormat(self, groupsep, decimalsep):
        changed = False

        if groupsep != self.grouping_sep_edit_cb.text():
            self.grouping_sep_edit_cb.setText(groupsep)
            changed = True

        if decimalsep != self.grouping_sep_edit_cb.text():
            self.decimal_sep_edit_cb.setText(decimalsep)
            changed = True

        if changed:
            self.__update_numbers_format()
            self.optionsChanged.emit()

    def numbersFormat(self):
        group = self.grouping_sep_edit_cb.text()
        decimal = self.decimal_sep_edit_cb.text()
        return {"group": group, "decimal": decimal}

    def __decimal_sep_activated(self, sep):
        group_sep = self.grouping_sep_edit_cb.text()
        preferred_replace = {".": ",", ",": "."}
        if sep == group_sep and sep in preferred_replace:
            self.grouping_sep_edit_cb.setText(preferred_replace[sep])
        elif sep == group_sep:
            cb = self.grouping_sep_edit_cb
            cb.setCurrentIndex((cb.currentIndex() + 1) % cb.count())

        self.__update_numbers_format()
        self.optionsEdited.emit()
        self.optionsChanged.emit()

    def __group_sep_activated(self, sep):
        decimal_sep = self.decimal_sep_edit_cb.text()
        preferred_replace = {".": ",", ",": "."}
        if sep == decimal_sep and sep in preferred_replace:
            self.decimal_sep_edit_cb.setText(preferred_replace[sep])
        elif sep == decimal_sep:
            cb = self.decimal_sep_edit_cb
            cb.setCurrentIndex((cb.currentIndex() + 1) % cb.count())
        self.__update_numbers_format()

        self.optionsEdited.emit()
        self.optionsChanged.emit()

    def __update_numbers_format(self):
        groupsep = self.grouping_sep_edit_cb.text()
        decimalsep = self.decimal_sep_edit_cb.text()

        model = self.__previewmodel
        if model is None:
            return
        parser = number_parser(groupsep, decimalsep)

        # update the delegates
        view = self.dataview
        for i in range(model.columnCount()):
            coltype = model.headerData(
                i, Qt.Horizontal, TablePreviewModel.ColumnTypeRole)
            if coltype == ColumnType.Numeric:
                delegate = ColumnValidateItemDelegate(view, converter=parser)
                view.setItemDelegateForColumn(i, delegate)

    def columnTypes(self):
        # type: () -> Dict[int, ColumnType]
        """
        Return the current column type annotations.

        Returns
        -------
        mapping : Dict[int, Optional[ColumnType]]
            Mapping from column indices column types.
        """
        # types = dict.fromkeys(range(model.rowCount()), ColumnType.Auto)
        types = {}
        types.update(self.__columnTypes())
        return types

    def setColumnTypes(self, types):
        # type: (Dict[int, Optional[ColumnType]]) -> None
        """
        Set column type annotations.

        Parameters
        ----------
        types : Dict[int, Optional[ColumnType]]
            Mapping from column indices to column types, `None` indicates
            default (unspecified type, will be inferred)
        """
        # This depends on encoding/dialect. Force preview model update.
        if self.__timer.isActive():
            self.__resetPreview()
        self.__setColumnTypes(types)

    def setStateForRow(self, row, state):
        # type: (int, TablePreview.RowSpec) -> None
        """
        Set the state for row.
        """
        if self.__timer.isActive():
            self.__resetPreview()

        model = self.__previewmodel
        if model is None:
            return
        rowcount = model.rowCount()
        while row > rowcount - 1 and model.canFetchMore():
            model.fetchMore()
            if model.rowCount() == rowcount:
                break
            rowcount = model.rowCount()

        model.setHeaderData(
            row, Qt.Vertical, state, TablePreviewModel.RowStateRole)
        self.dataview.setRowHints({row: state})

    def stateForRow(self, row):
        # type: (int) -> Optional[TablePreview.RowSpec]
        """
        Return the state for row.
        """
        model = self.__previewmodel
        if model is not None:
            return model.headerData(
                row, Qt.Vertical, TablePreviewModel.RowStateRole)
        else:
            return None

    def rowStates(self):
        # type: () -> (Dict[int, RowSpec])
        """
        Return states for all rows with non None state
        """
        return self.__rowStates()

    def setRowStates(self, rowstate):
        # type: (Dict[int, RowSpec]) -> None
        """
        Set the state for rows.

        Note
        ----
        States for all rows not passed in rowstate is reset to `None`.
        """
        if self.__timer.isActive():
            self.__resetPreview()
        model = self.__previewmodel
        if model is None:
            return

        currstate = self.rowStates()
        newstate = dict.fromkeys(currstate.keys(), None)
        newstate.update(rowstate)
        for row, state in newstate.items():
            self.setStateForRow(row, state)

    def __rowStates(self):
        model = self.__previewmodel
        items = (
            (row, model.headerData(row, Qt.Vertical,
                                   TablePreviewModel.RowStateRole))
            for row in range(model.rowCount())
        )
        return {row: state for row, state in items if state is not None}

    def setSampleContents(self, stream):
        # type: (io.BinaryIO) -> None
        """
        Set a binary file-like stream for displaying sample content.

        The stream will be read as needed when the data view is scrolled.

        Note
        ----
        If the stream is not seekable, its contents will be cached in memory.
        If and existing stream is already set it is NOT closed. The caller
        is responsible for managing its lifetime.
        """
        self.__sample = stream
        self.__buffer = io.BytesIO()
        self.__resetPreview()

    def __resetPreview(self):
        # Reset the preview model and view
        self.__timer.stop()
        colstate = {}
        rowstate = {}
        if self.__previewmodel is not None:
            # store the column/row specs
            colstate = self.__columnTypes()
            rowstate = self.__rowStates()
            self.__previewmodel.errorOccurred.disconnect(self.__set_error)
            self.__previewmodel.deleteLater()
            self.__previewmodel = None

        if self.__textwrapper is not None:
            self.__textwrapper.detach()
            self.__textwrapper = None

        self.__set_error("")
        self.previewModelReset.emit()

        if self.__sample is None:
            return

        self.__previewmodel = TablePreviewModel(self)
        self.__previewmodel.errorOccurred.connect(self.__set_error)
        try:
            seekable = self.__sample.seekable()
        except AttributeError:
            seekable = False

        if seekable:
            # Might be better to always use buffer? (compressed streams are
            # seekable but slower)
            base = self.__sample
            base.seek(0)
        else:
            self.__buffer.seek(0)
            base = CachedBytesIOWrapper(self.__sample, self.__buffer)

        wrapper = io.TextIOWrapper(
            base, encoding=self.encoding(),
            # use surrogate escape to validate/detect encoding errors in
            # delegates
            errors="surrogateescape"
        )
        rows = csv.reader(
            wrapper, dialect=self.dialect()
        )

        self.__textwrapper = wrapper
        self.__previewmodel.setPreviewStream(rows)
        if self.__previewmodel.canFetchMore():
            # TODO: Fetch until the same number of rows as at method entry?
            self.__previewmodel.fetchMore()

        self.dataview.setModel(self.__previewmodel)
        self.dataview.selectionModel().selectionChanged.connect(
            self.__update_column_type_edit, Qt.UniqueConnection
        )
        if self.__previewmodel.columnCount() == len(colstate):
            self.__setColumnTypes(colstate)
        for row, state in rowstate.items():
            self.__previewmodel.setHeaderData(row, Qt.Vertical, state,
                                              TablePreviewModel.RowStateRole)
        self.dataview.setRowHints(rowstate)

    @Slot()
    def __update_column_type_edit(self):
        # Update the 'Column type' edit control based on current column
        # selection
        smodel = self.dataview.selectionModel()
        model = self.dataview.model()
        cb = self.column_type_edit_cb
        columns = smodel.selectedColumns(0)
        types = {model.headerData(c.column(), Qt.Horizontal,
                                  TablePreviewModel.ColumnTypeRole)
                 for c in columns}

        types = {ColumnType.Auto if t is None else t for t in types}
        if len(types) == 0:
            # no selection, disabled
            cb.setCurrentIndex(-1)
            cb.setEnabled(False)
        elif len(types) == 1:
            idx = cb.findData(types.pop(), Qt.UserRole)
            cb.setCurrentIndex(idx)
            cb.setEnabled(True)
        else:
            cb.setCurrentIndex(-1)
            cb.setEnabled(True)

    def __on_column_type_edit_activated(self, idx):
        # Column type set via the combo box.
        coltype = self.column_type_edit_cb.itemData(idx, Qt.UserRole)
        smodel = self.dataview.selectionModel()
        columns = smodel.selectedColumns(0)
        columns = [c.column() for c in columns]
        self.__setColumnType(columns, coltype)

    def __dataview_context_menu(self, pos):
        pos = self.dataview.viewport().mapToGlobal(pos)
        cols = self.dataview.selectionModel().selectedColumns(0)
        cols = [idx.column() for idx in cols]
        self.__run_type_columns_menu(pos, cols)

    def __hheader_context_menu(self, pos):
        pos = self.dataview.horizontalHeader().mapToGlobal(pos)
        cols = self.dataview.selectionModel().selectedColumns(0)
        cols = [idx.column() for idx in cols]
        self.__run_type_columns_menu(pos, cols)

    def __vheader_context_menu(self, pos):
        header = self.dataview.verticalHeader()  # type: QHeaderView
        index = header.logicalIndexAt(pos)
        pos = header.mapToGlobal(pos)
        model = header.model()  # type: QAbstractTableModel

        RowStateRole = TablePreviewModel.RowStateRole
        state = model.headerData(index, Qt.Vertical, RowStateRole)
        m = QMenu(header)
        skip_action = m.addAction("Skip")
        skip_action.setCheckable(True)
        skip_action.setChecked(state == TablePreview.Skipped)
        m.addSection("")
        mark_header = m.addAction("Header")
        mark_header.setCheckable(True)
        mark_header.setChecked(state == TablePreview.Header)

        def update_row_state(action):
            # type: (QAction) -> None
            state = None
            if action is mark_header:
                state = TablePreview.Header if action.isChecked() else None
            elif action is skip_action:
                state = TablePreview.Skipped if action.isChecked() else None
            model.setHeaderData(index, Qt.Vertical, state, RowStateRole)
            self.dataview.setRowHints({index: state})

        m.triggered.connect(update_row_state)
        m.popup(pos)

    def __run_type_columns_menu(self, pos, columns):
        # type: (QPoint, List[int]) -> None
        # Open a QMenu at pos for setting column types for column indices list
        # `columns`
        model = self.__previewmodel
        if model is None:
            return
        menu = QMenu(self)
        menu.setAttribute(Qt.WA_DeleteOnClose)
        coltypes = {model.headerData(
                        i, Qt.Horizontal, TablePreviewModel.ColumnTypeRole)
                    for i in columns}
        coltypes = {ColumnType.Auto if t is None else t for t in coltypes}
        if len(coltypes) == 1:
            current = coltypes.pop()
        else:
            current = None
        cb = self.column_type_edit_cb
        g = QActionGroup(menu)
        current_action = None
        # 'Copy' the column types model into a menu
        for i in range(cb.count()):
            if cb.itemData(i, Qt.AccessibleDescriptionRole) == "separator":
                menu.addSeparator()
                continue

            ac = menu.addAction(cb.itemIcon(i), cb.itemText(i))
            ac.setData(cb.itemData(i, Qt.UserRole))
            ac.setCheckable(True)
            if ac.data() == current:
                ac.setChecked(True)
                current_action = ac
            g.addAction(ac)

        def update_types(action):
            newtype = action.data()
            self.__setColumnType(columns, newtype)

        menu.triggered.connect(update_types)
        menu.triggered.connect(self.__update_column_type_edit)
        menu.popup(pos, current_action)

    def __setColumnType(self, columns, coltype):
        # type: (List[int], ColumnType) -> None
        view = self.dataview
        model = view.model()  # type: QAbstractTableModel
        numbersformat = self.numbersFormat()
        numberconverter = number_parser(
            numbersformat["group"], numbersformat["decimal"])
        if coltype == ColumnType.Numeric:
            delegate = ColumnValidateItemDelegate(self.dataview,
                                                  converter=numberconverter)
        elif coltype == ColumnType.Text:
            delegate = ColumnValidateItemDelegate(self.dataview,
                                                  converter=str.strip)
        elif coltype == ColumnType.Time:
            delegate = ColumnValidateItemDelegate(self.dataview,
                                                  converter=parse_datetime)
        elif coltype == ColumnType.Skip:
            delegate = SkipItemDelegate(self.dataview)
        else:
            delegate = None

        changed = False
        for i in columns:
            current = model.headerData(
                i, Qt.Horizontal, TablePreviewModel.ColumnTypeRole)
            changed = changed or current != coltype
            model.setHeaderData(
                i, Qt.Horizontal, coltype, TablePreviewModel.ColumnTypeRole
            )
            model.setHeaderData(
                i, Qt.Horizontal, icon_for_column_type(coltype),
                Qt.DecorationRole
            )
            self.dataview.setItemDelegateForColumn(i, delegate)

        if changed:
            self.__update_column_type_edit()
            self.columnTypesChanged.emit()

    def __setColumnTypes(self, coltypes):
        # type: (Dict[int, ColumnType]) -> None
        def mapping_invert(mapping):
            # type: (Dict[_A, _B]) -> Dict[_B, List[_A]]
            m = defaultdict(list)
            for key, val in mapping.items():
                m[val].append(key)
            return m
        model = self.__previewmodel
        if model is None:
            return
        coltypes_ = dict.fromkeys(range(model.columnCount()), ColumnType.Auto)
        coltypes_.update(coltypes)
        for coltype, cols in mapping_invert(coltypes_).items():
            self.__setColumnType(cols, coltype)

    def __columnTypes(self):
        # type: () -> Dict[int, ColumnType]
        model = self.__previewmodel
        if model is None:
            return {}
        res = {
            i: model.headerData(i, Qt.Horizontal,
                                TablePreviewModel.ColumnTypeRole)
            for i in range(model.columnCount())
        }
        return {i: val for i, val in res.items()
                if val is not None and val != ColumnType.Auto}

    def columnTypeRanges(self):
        # type: () -> List[Tuple[range, ColumnType]]
        """
        Return the column type specs as column ranges.

        Returns
        -------
        coltypes : List[Tuple[range, ColumnType]]
            A list of `(range, coltype)` tuples where `range` are ranges
            with step 1 and coltype a ColumnType. The ranges are sorted
            in ascending order.

        Note
        ----
        Unlike `columnTypes` this method does not omit ColumnTypes.Auto
        entries.
        """
        model = self.__previewmodel
        if model is None:
            return []
        res = dict.fromkeys(range(model.columnCount()), ColumnType.Auto)
        res.update(self.__columnTypes())
        types = sorted(res.items())
        res = []

        # Group by increasing indices (with step 1) and coltype
        def groupkey(item, __incseq=iter(itertools.count())):
            index, val = item
            return index - next(__incseq), val

        for (_, key), items in itertools.groupby(types, key=groupkey):
            items = list(items)
            start = items[0][0]
            last = items[-1][0]
            res.append((range(start, last + 1), key))
        return res

    def setColumnTypeRanges(self, ranges):
        # type: (List[Tuple[range, ColumnType]]) -> None
        """
        Set column type specs for ranges.

        Parameters
        ----------
        ranges : List[Tuple[range, ColumnType]]
            For every `(range, coltype)` tuple set the corresponding coltype.
        """
        self.setColumnTypes({i: coltype for r, coltype in ranges for i in r})

    def __set_error(self, errorstr):
        # type: (str) -> None
        if not errorstr:
            self.__overlay.hide()
            self.__error_label.setText("")
        else:
            self.__overlay.show()
            self.__error_label.setText(errorstr)
            self.previewReadErrorOccurred.emit(errorstr)


class CachedBytesIOWrapper(io.BufferedIOBase):
    """
    Read and cache data from `base`. When cache is not empty prepend data from
    the cache before switching to base

    Base needs to implement `read` method, cache must be read/write and
    seekable.

    Utility wrapper to implement restartable reads for streams that are not
    seekable.
    """
    def __init__(self, base, cache):
        # type: (io.BinaryIO, io.BytesIO) -> None
        super().__init__()
        self.__base = base
        self.__cache = cache

    def detach(self):
        base = self.__base
        self.__base = None
        return base

    def read(self, size=-1):
        # type: (Optional[int]) -> bytes
        base, cache = self.__base, self.__cache
        if size is None or size < 0:
            b1 = cache.read()
            b2 = base.read()
            cache.write(b2)
            return b1 + b2
        else:
            if cache.tell() < len(cache.getbuffer()):
                b1 = cache.read(size)
                if len(b1) < size:
                    assert len(cache.getbuffer()) == cache.tell()
                    b2 = base.read(size - len(b1))
                    cache.write(b2)
                    assert len(cache.getbuffer()) == cache.tell()
                    b = b1 + b2
                else:
                    b = b1
            else:
                b = base.read(size)
                cache.write(b)
                assert len(cache.getbuffer()) == cache.tell()
            return b

    def read1(self, size=-1):
        # Does not exactly conform to spec, but necessary for io.TextIOWrapper
        return self.read(size)

    def readable(self):
        return True

    def writable(self):
        return False


class RowSpec(enum.IntEnum):
    """Row spec flags"""
    #: Header row
    Header = 1
    #: Row is skipped
    Skipped = 2


class TablePreview(QTableView):
    RowSpec = RowSpec
    Header, Skipped = RowSpec

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setItemDelegate(PreviewItemDelegate(self))

    def rowsInserted(self, parent, start, end):
        # type: (QModelIndex, int, int) -> None
        super().rowsInserted(parent, start, end)
        behavior = self.selectionBehavior()
        if behavior in (QTableView.SelectColumns, QTableView.SelectRows):
            # extend the selection to the new rows
            smodel = self.selectionModel()
            selection = smodel.selection()
            command = QItemSelectionModel.Select
            if behavior == QTableView.SelectRows:
                command |= QItemSelectionModel.Rows
            if behavior == QTableView.SelectColumns:
                command |= QItemSelectionModel.Columns
            smodel.select(selection, command)

    def setRowHints(self, hints):
        # type: (Dict[int, TablePreview.RowSpec]) -> None
        for row, hint in hints.items():
            current = self.itemDelegateForRow(row)
            if current is not None:
                current.deleteLater()
            if hint == TablePreview.Header:
                delegate = HeaderItemDelegate(self)
            elif hint == TablePreview.Skipped:
                delegate = SkipItemDelegate(self)
            else:
                delegate = None
            self.setItemDelegateForRow(row, delegate)

    def sizeHint(self):
        sh = super().sizeHint()  # type: QSize
        hh = self.horizontalHeader()  # type: QHeaderView
        vh = self.verticalHeader()  # type: QHeaderView
        hsection = hh.defaultSectionSize()
        vsection = vh.defaultSectionSize()
        return sh.expandedTo(QSize(8 * hsection, 20 * vsection))


def is_surrogate_escaped(text: str) -> bool:
    """Does `text` contain any surrogate escape characters."""
    return any("\udc80" <= c <= "\udcff" for c in text)


class PreviewItemDelegate(QStyledItemDelegate):
    def initStyleOption(self, option, index):
        # type: (QStyleOptionViewItem, QModelIndex) -> None
        super().initStyleOption(option, index)
        if len(option.text) > 500:
            # Shorten long text (long text layout takes too long)
            f = QTextBoundaryFinder(QTextBoundaryFinder.Grapheme, option.text)
            f.setPosition(500)
            i = f.toNextBoundary()
            if i != -1:
                option.text = option.text[:i] + "..."
        model = index.model()
        coltype = model.headerData(index.column(), Qt.Horizontal,
                                   TablePreviewModel.ColumnTypeRole)
        if coltype == ColumnType.Numeric or coltype == ColumnType.Time:
            option.displayAlignment = Qt.AlignRight | Qt.AlignVCenter

        if not self.validate(option.text):
            option.palette.setBrush(
                QPalette.All, QPalette.Text, QBrush(Qt.red, Qt.SolidPattern)
            )
            option.palette.setBrush(
                QPalette.All, QPalette.HighlightedText,
                QBrush(Qt.red, Qt.SolidPattern)
            )

    def validate(self, value: str) -> bool:  # pylint: disable=no-self-use
        return not is_surrogate_escaped(value)

    def helpEvent(self, event, view, option, index):
        # type: (QHelpEvent, QAbstractItemView, QStyleOptionViewItem, QModelIndex) -> bool
        if event.type() == QEvent.ToolTip:
            ttip = index.data(Qt.ToolTipRole)
            if ttip is None:
                ttip = index.data(Qt.DisplayRole)
                ttip = self.displayText(ttip, option.locale)
                QToolTip.showText(event.globalPos(), ttip, view)
                return True
        return super().helpEvent(event, view, option, index)


class HeaderItemDelegate(PreviewItemDelegate):
    """
    Paint the items with an alternate color scheme
    """
    NoFeatures = 0
    AutoDecorate = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__features = HeaderItemDelegate.NoFeatures

    def features(self):
        return self.__features

    def initStyleOption(self, option, index):
        # type: (QStyleOptionViewItem, QModelIndex) -> None
        super().initStyleOption(option, index)
        palette = option.palette
        shadow = palette.color(QPalette.WindowText)  # type: QColor
        if shadow.isValid():
            shadow.setAlphaF(0.1)
            option.backgroundBrush = QBrush(shadow, Qt.SolidPattern)
        option.displayAlignment = Qt.AlignCenter
        model = index.model()
        if option.icon.isNull() and \
                self.__features & HeaderItemDelegate.AutoDecorate:
            ctype = model.headerData(index.column(), Qt.Horizontal,
                                     TablePreviewModel.ColumnTypeRole)
            option.icon = icon_for_column_type(ctype)
        if not option.icon.isNull():
            option.features |= QStyleOptionViewItem.HasDecoration


def icon_for_column_type(coltype):
    # type: (ColumnType) -> QIcon
    if coltype == ColumnType.Numeric:
        icon = QIcon(StampIconEngine("N", QColor("red")))
    elif coltype == ColumnType.Categorical:
        icon = QIcon(StampIconEngine("C", QColor("green")))
    elif coltype == ColumnType.Text:
        icon = QIcon(StampIconEngine("S", QColor("black")))
    elif coltype == ColumnType.Time:
        icon = QIcon(StampIconEngine("T", QColor("deepskyblue")))
    else:
        icon = QIcon()
    return icon


class SkipItemDelegate(PreviewItemDelegate):
    def initStyleOption(self, option, index):
        # type: (QStyleOptionViewItem, QModelIndex) -> None
        super().initStyleOption(option, index)
        color = QColor(Qt.red)
        palette = option.palette  # type: QPalette
        base = palette.color(QPalette.Base)
        if base.isValid() and base.value() > 127:
            # blend on 'light' base, not on dark (low contrast)
            color.setAlphaF(0.2)
        option.backgroundBrush = QBrush(color, Qt.DiagCrossPattern)


class ColumnValidateItemDelegate(PreviewItemDelegate):
    def __init__(self, *args, converter=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.converter = converter or float

    def validate(self, value):
        if value in {"NA", "Na", "na", "n/a", "N/A", "?", "", "."}:
            return True
        try:
            self.converter(value)
        except ValueError:
            return False
        else:
            return super().validate(value)


def number_parser(groupsep, decimalsep):
    # type: (str, str) -> Callable[[str], float]
    if groupsep == "" and decimalsep == ".":
        return float
    elif groupsep == "":
        return lambda value: float(value.replace(decimalsep, "."))
    elif decimalsep != groupsep and decimalsep != "" and groupsep != "":
        table = {ord(groupsep): None, ord(decimalsep): ord(".")}
        return lambda value: float(value.translate(table))
    else:
        return float


class TablePreviewModel(QAbstractTableModel):
    """
    Lazy populated table preview model.

    The model reads rows on demand from an 'rows' iterable when requested
    (via fetchMore).
    Additionally the client can set column/row header data.
    """
    ColumnTypeRole = Qt.UserRole + 11
    RowStateRole = Qt.UserRole + 12

    #: Signal emitted when an error occurs while iterating over the preview
    #: stream.
    errorOccurred = Signal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__rowCount = self.__colCount = 0
        self.__rows = []
        self.__canFetchMore = False
        self.__error = None
        self.__iter = None
        # extra header data for use by setHeaderData
        self.__headerData = {
            Qt.Horizontal: defaultdict(dict),
            Qt.Vertical: defaultdict(dict),
        }

    def setPreviewStream(self, stream):
        # type: (Iterator[List[str]]) -> None
        """
        Set an iterator over the rows.

        The iterator will be advanced on demand by `fetchMore`, while storing
        the returned values. Previous stream and its cached data is discarded.
        """
        self.beginResetModel()
        self.__iter = stream
        self.__rows = []
        self.__rowCount = self.__colCount = 0
        self.__canFetchMore = True
        self.__error = None
        self.endResetModel()

    def canFetchMore(self, parent=QModelIndex()):
        """Reimplemented."""
        if not parent.isValid():
            return self.__canFetchMore
        else:
            return False

    def fetchMore(self, parent=QModelIndex()):
        """Reimplemented."""
        if not parent.isValid():
            error = self.__error
            if self.__rowCount == 0:
                newrows = self.__tryFetchRows(20)
            else:
                newrows = self.__tryFetchRows(5)

            if newrows:
                extent = len(newrows), max(len(row) for row in newrows)
                rows, cols = self.__rowCount, self.__colCount

                self.beginInsertRows(QModelIndex(), rows, rows + extent[0] - 1)
                self.__rows.extend(newrows)
                self.__rowCount += extent[0]
                self.endInsertRows()

                if cols < extent[1]:
                    newColCount = max(cols, extent[1])
                    self.beginInsertColumns(QModelIndex(), cols, newColCount - 1)
                    self.__colCount = newColCount
                    self.endInsertColumns()

            # Emit error after inserting the final rows
            if self.__error is not None and self.__error != error:
                self.errorOccurred.emit(self.__error)

    def __tryFetchRows(self, n=10):
        # type: (int) -> List[List[str]]
        """
        Fetch and return a maximum of `n` rows from the source preview stream.
        """
        rows = []
        for _ in range(n):
            try:
                row = next(self.__iter)
            except StopIteration:
                self.__canFetchMore = False
                break
            except Exception as err:  # pylint: disable=broad-except
                print("".join(traceback.format_exception(*sys.exc_info())),
                      file=sys.stderr)
                self.__error = format_exception(err)
                self.__canFetchMore = False
                break
            else:
                rows.append(row)
        return rows

    def rowCount(self, parent=QModelIndex()):
        # type: (QModelIndex) -> int
        """Reimplemented."""
        return 0 if parent.isValid() else self.__rowCount

    def columnCount(self, parent=QModelIndex()):
        # type: (QModelIndex) -> int
        """Reimplemented."""
        return 0 if parent.isValid() else self.__colCount

    def data(self, index, role=Qt.DisplayRole):
        # type: (QModelIndex, int) -> Any
        """Reimplemented."""
        if not index.isValid():
            return None
        row, col = index.row(), index.column()
        assert self.__rowCount == len(self.__rows)
        if not 0 <= row < self.__rowCount:
            return None
        row = self.__rows[row]

        if not 0 <= col < len(row):
            return None

        value = row[col]
        if role == Qt.DisplayRole:
            return value
        elif role == TablePreviewModel.ColumnTypeRole:
            return self.__headerData[Qt.Horizontal][index.column()].get(role)
        else:
            return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        # type: (int, Qt.Orientation, int) -> Any
        """Reimplemented."""
        if role == Qt.DisplayRole:
            return section + 1
        else:
            return self.__headerData[orientation][section].get(role)

    def setHeaderData(self, section, orientation, value, role=Qt.EditRole):
        # type: (int, Qt.Orientation, Any, Qt.ItemDataRole) -> bool
        """Reimplemented."""
        current = self.__headerData[orientation][section].get(role, None)
        if current != value:
            if value is None:
                del self.__headerData[orientation][section][role]
            else:
                self.__headerData[orientation][section][role] = value
            self.headerDataChanged.emit(orientation, section, section)
        return True

    def updateHeaderData(self, orientation, values):
        # type: (Qt.Orientation, Dict[int, Dict[Qt.ItemDataRole, Any]]) -> None
        """
        Update/set multiple header sections/roles at once.

        Parameters
        ----------
        orientation : Qt.Orientation
        values : Dict[int, Dict[Qt.ItemDataRole, Any]]
            A mapping of section indices to mapping of role to values.
            e.g. `{1: {Qt.DisplayRole: "A"}}` sets the display text to "A"
        """
        data = self.__headerData[orientation]
        if orientation == Qt.Horizontal:
            length = self.__colCount
        else:
            length = self.__rowCount
        sections = []
        for section, itemdata in values.items():
            if 0 <= section < length:
                data[section].update(itemdata)
                sections.append(section)
        if not sections:
            return

        first = min(sections)
        last = max(sections)

        self.headerDataChanged.emit(orientation, first, last)

    def flags(self, index):
        # type: (QModelIndex) -> Qt.ItemFlags
        """Reimplemented."""
        # pylint: disable=unused-argument,no-self-use
        return Qt.ItemFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)

    def errorString(self):
        # type: () -> Optional[str]
        """
        Return the error string or None if no error occurred.
        """
        return self.__error


@singledispatch
def format_exception(err):
    return "".join(traceback.format_exception_only(type(err), err)).rstrip()


@format_exception.register(csv.Error)
def format_exception_csv(err):
    return "CSV parsing error: " + str(err)


_to_datetime = None


def parse_datetime(text):
    global _to_datetime
    if _to_datetime is None:
        from pandas import to_datetime as _to_datetime
    return _to_datetime(text)


TEST_DATA = b"""\
 ,A,B,C,D
1,a,1,1,
2,b,2,2,
3,c,3,3,
4,d,4,4,,\
"""


def main(argv=None):  # pragma: no cover
    app = QApplication(argv or [])
    argv = app.arguments()
    w = CSVImportWidget()
    w.show()
    w.raise_()

    if len(argv) > 1:
        path = argv[1]
        f = open(path, "rb")
    else:
        f = io.BytesIO(TEST_DATA)
    try:
        w.setSampleContents(f)
        app.exec()
    finally:
        f.close()


if __name__ == "__main__":  # pragma: no cover
    csv.field_size_limit(4 * 2 ** 20)
    main(sys.argv)



# Pull members from modules to Orange.modelling namespace
# pylint: disable=wildcard-import

from .base import *

from .ada_boost import *
from .constant import *
from .knn import *
from .linear import *
from .neural_network import *
from .randomforest import *
from .svm import *
from .tree import *
try:
    from .catgb import *
except ImportError:
    pass
from .gb import *
try:
    from .xgb import *
except ImportError:
    pass



from hashlib import sha1

import numpy as np

from Orange.classification import Learner, Model
from Orange.statistics import distribution

__all__ = ["MajorityLearner"]


class MajorityLearner(Learner):
    """
    A majority classifier. Always returns most frequent class from the
    training set, regardless of the attribute values from the test data
    instance. Returns class value distribution if class probabilities
    are requested. Can be used as a baseline when comparing classifiers.

    In the special case of uniform class distribution within the training data,
    class value is selected randomly. In order to produce consistent results on
    the same dataset, this value is selected based on hash of the class vector.
    """
    def fit_storage(self, dat):
        if not dat.domain.has_discrete_class:
            raise ValueError("classification.MajorityLearner expects a domain "
                             "with a (single) categorical variable")
        dist = distribution.get_distribution(dat, dat.domain.class_var)
        N = dist.sum()
        if N > 0:
            dist /= N
        else:
            dist.fill(1 / len(dist))

        probs = np.array(dist)
        ties = np.flatnonzero(probs == probs.max())
        if len(ties) > 1:
            random_idx = int(sha1(np.ascontiguousarray(dat.Y).data)
                             .hexdigest(), 16) % len(ties)
            unif_maj = ties[random_idx]
        else:
            unif_maj = None
        return ConstantModel(dist=dist, unif_maj=unif_maj)


class ConstantModel(Model):
    """
    A classification model that returns a given class value.
    """
    def __init__(self, dist, unif_maj=None):
        """
        Constructs `Orange.classification.MajorityModel` that always
        returns majority value of given distribution.

        If no or empty distribution given, constructs a model that returns equal
        probabilities for each class value.

        :param dist: domain for the `Table`
        :param unif_maj: majority class for the special case of uniform
            class distribution in the training data
        :type dist: Orange.statistics.distribution.Discrete
        :return: regression model that returns majority value
        :rtype: Orange.classification.Model
        """
        self.dist = np.array(dist)
        self.unif_maj = unif_maj

    def predict(self, X):
        """
        Returns majority class for each given instance in X.

        :param X: data table for which to make predictions
        :type X: Orange.data.Table
        :return: predicted value
        :rtype: vector of majority values
        """
        probs = np.tile(self.dist, (X.shape[0], 1))
        if self.unif_maj is not None:
            value = np.tile(self.unif_maj, (X.shape[0], ))
            return value, probs
        return probs

    def __str__(self):
        return 'ConstantModel {}'.format(self.dist)

MajorityLearner.__returns__ = ConstantModel



from AnyQt.QtCore import Qt, QEvent
from AnyQt.QtGui import QFont, QColor, QFocusEvent
from AnyQt.QtWidgets import QApplication
from AnyQt.QtTest import QTest, QSignalSpy

from orangewidget.tests.base import GuiTest
from orangewidget.tests.utils import simulate
from orangewidget.utils.itemmodels import PyListModel
from Orange.widgets.utils.combobox import ItemStyledComboBox, TextEditCombo


class TestItemStyledComboBox(GuiTest):
    def test_combobox(self):
        cb = ItemStyledComboBox()
        cb.setPlaceholderText("...")
        self.assertEqual(cb.placeholderText(), "...")
        cb.grab()
        cb.addItems(["1"])
        cb.setCurrentIndex(0)
        model = cb.model()
        model.setItemData(model.index(0, 0), {
            Qt.ForegroundRole: QColor(Qt.blue),
            Qt.FontRole: QFont("Windings")
        })
        cb.grab()


class TestTextEditCombo(GuiTest):
    def test_texteditcombo(self):
        cb = TextEditCombo()
        model = PyListModel()
        cb.setModel(model)

        def enter_text(text: str):
            cb.lineEdit().selectAll()
            spy_act = QSignalSpy(cb.activated[int])
            spy_edit = QSignalSpy(cb.editingFinished)
            QTest.keyClick(cb.lineEdit(), Qt.Key_Delete)
            QTest.keyClicks(cb.lineEdit(), text)
            QApplication.sendEvent(
                cb, QFocusEvent(QEvent.FocusOut, Qt.TabFocusReason)
            )
            self.assertEqual(len(spy_edit), 1)
            if cb.insertPolicy() != TextEditCombo.NoInsert:
                self.assertEqual(list(spy_act), [[cb.currentIndex()]])

        cb.setInsertPolicy(TextEditCombo.NoInsert)
        enter_text("!!")
        self.assertEqual(list(model), [])
        cb.setInsertPolicy(TextEditCombo.InsertAtTop)
        enter_text("BB")
        enter_text("AA")
        self.assertEqual(list(model), ["AA", "BB"])
        cb.setInsertPolicy(TextEditCombo.InsertAtBottom)
        enter_text("CC")
        self.assertEqual(list(model), ["AA", "BB", "CC"])
        cb.setInsertPolicy(TextEditCombo.InsertBeforeCurrent)
        cb.setCurrentIndex(1)
        enter_text("AB")
        self.assertEqual(list(model), ["AA", "AB", "BB", "CC"])
        cb.setInsertPolicy(TextEditCombo.InsertAfterCurrent)
        cb.setCurrentIndex(2)
        enter_text("BC")
        self.assertEqual(list(model), ["AA", "AB", "BB", "BC", "CC"])
        cb.setInsertPolicy(TextEditCombo.InsertAtCurrent)
        cb.setCurrentIndex(2)
        enter_text("BBA")
        self.assertEqual(list(model), ["AA", "AB", "BBA", "BC", "CC"])
        cb.setInsertPolicy(TextEditCombo.InsertAlphabetically)
        enter_text("BCA")
        self.assertEqual(list(model), ["AA", "AB", "BBA", "BC", "BCA", "CC"])

    def test_activate_editing_finished_emit_ordering(self):
        def activated():
            sigs.append("activated")

        def finished():
            sigs.append("finished")

        sigs = []
        cb = TextEditCombo(
            activated=activated, editingFinished=finished
        )
        cb.insertItem(0, "AA")
        simulate.combobox_activate_index(cb, 0)
        self.assertEqual(sigs, ["finished", "activated"])



from Orange.classification import SklTreeLearner
from Orange.classification import TreeLearner as ClassificationTreeLearner
from Orange.modelling import Fitter, SklFitter
from Orange.regression import TreeLearner as RegressionTreeLearner
from Orange.regression.tree import SklTreeRegressionLearner
from Orange.tree import TreeModel

__all__ = ['SklTreeLearner', 'TreeLearner']


class SklTreeLearner(SklFitter):
    name = 'tree'

    __fits__ = {'classification': SklTreeLearner,
                'regression': SklTreeRegressionLearner}


class TreeLearner(Fitter):
    name = 'tree'

    __fits__ = {'classification': ClassificationTreeLearner,
                'regression': RegressionTreeLearner}

    __returns__ = TreeModel
